# -*- coding: utf-8 -*-

"""
Here we define jack readers. Jack readers consist of 3 layers, one that transform
jack data structures into tensors, one that processes predicts the outputs and losses
using a TensorFlow model into other tensors, and one that converts these tensors back to jack data structures.
"""

import logging
import math
import os
import shutil
from typing import Iterable, List

import progressbar

from jack.core.data_structures import *
from jack.core.input_module import InputModule
from jack.core.model_module import ModelModule
from jack.core.output_module import OutputModule
from jack.core.shared_resources import SharedResources

logger = logging.getLogger(__name__)


class JTReader:
    """
    A tensorflow reader reads inputs consisting of questions, supports and possibly candidates, and produces answers.
    It consists of three layers: input to tensor (input_module), tensor to tensor (model_module), and tensor to answer
    (output_model). These layers are called in-turn on a given input (list).
    """

    def __init__(self,
                 shared_resources: SharedResources,
                 input_module: InputModule,
                 model_module: ModelModule,
                 output_module: OutputModule):
        self._shared_resources = shared_resources
        self._output_module = output_module
        self._model_module = model_module
        self._input_module = input_module
        self._is_setup = False

        assert all(port in self.input_module.output_ports for port in self.model_module.input_ports), \
            "Input Module outputs must include model module inputs"

        assert all(port in self.input_module.training_ports or port in self.model_module.output_ports or
                   port in self.input_module.output_ports for port in self.model_module.training_input_ports), \
            "Input Module (training) outputs and model module outputs must include model module training inputs"

        assert all(port in self.model_module.output_ports or port in self.input_module.output_ports
                   for port in self.output_module.input_ports), \
            "Module model output must match output module inputs"

    @property
    def input_module(self) -> InputModule:
        """Returns: input module"""
        return self._input_module

    @property
    def model_module(self) -> ModelModule:
        """Returns: model module"""
        return self._model_module

    @property
    def output_module(self) -> OutputModule:
        """Returns: output module"""
        return self._output_module

    @property
    def shared_resources(self) -> SharedResources:
        """Returns: SharedResources object"""
        return self._shared_resources

    def __call__(self, inputs: Sequence[QASetting]) -> Sequence[Answer]:
        """
        Answers a list of question settings
        Args:
            inputs: a list of inputs.

        Returns:
            predicted outputs/answers to a given (labeled) dataset
        """
        batch = self.input_module(inputs)
        output_module_input = self.model_module(batch, self.output_module.input_ports)
        answers = self.output_module(inputs, {p: output_module_input[p] for p in self.output_module.input_ports})
        return answers

    def process_dataset(self, dataset: Sequence[Tuple[QASetting, Answer]], batch_size: int, silent=True):
        """
        Similar to the call method, only that it works on a labeled dataset and applies batching. However, assumes
        that batches in input_module.batch_generator are processed in order and do not get shuffled during with
        flag is_eval set to true.

        Args:
            dataset:
            batch_size: note this information is needed here, but does not set the batch_size the model is using.
            This has to happen during setup/configuration.
            silent: if true, no output

        Returns:
            predicted outputs/answers to a given (labeled) dataset
        """
        batches = self.input_module.batch_generator(dataset, batch_size, is_eval=True)
        answers = list()
        enumerator = enumerate(batches)
        if not silent:
            logger.info("Start answering...")
            bar = progressbar.ProgressBar(
                max_value=math.ceil(len(dataset) / batch_size),
                widgets=[' [', progressbar.Timer(), '] ', progressbar.Bar(), ' (', progressbar.ETA(), ') '])
            enumerator = bar(enumerator)
        for j, batch in enumerator:
            output_module_input = self.model_module(batch, self.output_module.input_ports)
            questions = [q for q, a in dataset[j * batch_size:(j + 1) * batch_size]]
            answers.extend(a[0] for a in self.output_module(
                questions, {p: output_module_input[p] for p in self.output_module.input_ports}))

        return answers

    def train(self, optimizer, training_set: Iterable[Tuple[QASetting, List[Answer]]], batch_size: int,
              max_epochs=10, hooks=tuple(), **kwargs):
        """
        This method trains the reader (and changes its state).

        Args:
            optimizer: TF optimizer
            training_set: the training instances.
            max_epochs: maximum number of epochs
            hooks: TrainingHook implementations that are called after epochs and batches
            kwargs: additional reader specific options
        """
        raise NotImplementedError

    def setup_from_data(self, data: Iterable[Tuple[QASetting, List[Answer]]], is_training=False):
        """
        Sets up modules given a training dataset if necessary.

        Args:
            data: training dataset
            is_training: indicates whether it's the training phase or not
        """
        self.input_module.setup_from_data(data)
        self.input_module.setup()
        self.model_module.setup(is_training)
        self.output_module.setup()
        self._is_setup = True

    def load_and_setup(self, path, is_training=False):
        """
        Sets up already stored reader from model directory.

        Args:
            path: training dataset
            is_training: indicates whether it's the training phase or not
        """
        self.shared_resources.load(os.path.join(path, "shared_resources"))
        self.load_and_setup_modules(path, is_training)

    def load_and_setup_modules(self, path, is_training=False):
        """
        Sets up already stored reader from model directory.

        Args:
            path: training dataset
            is_training: indicates whether it's the training phase or not
        """
        self.input_module.setup()
        self.input_module.load(os.path.join(path, "input_module"))
        self.model_module.setup(is_training)
        self.model_module.load(os.path.join(path, "model_module"))
        self.output_module.setup()
        self.output_module.load(os.path.join(path, "output_module"))
        self._is_setup = True

    def load(self, path):
        """
        (Re)loads module states on a setup reader (but not shared resources).
        If reader is not setup yet use setup from file instead.

        Args:
            path: model directory
        """
        self.input_module.load(os.path.join(path, "input_module"))
        self.model_module.load(os.path.join(path, "model_module"))
        self.output_module.load(os.path.join(path, "output_module"))

    def store(self, path):
        """
        Store module states and shared resources.

        Args:
            path: model directory
        """
        if os.path.exists(path):
            shutil.rmtree(path)
        os.makedirs(path)
        self.shared_resources.store(os.path.join(path, "shared_resources"))
        self.input_module.store(os.path.join(path, "input_module"))
        self.model_module.store(os.path.join(path, "model_module"))
        self.output_module.store(os.path.join(path, "output_module"))



"""Shared resources are used to store reader all stateful information about a reader and share it between modules.

Examples are include the vocabulary, hyper-parameters or name of a reader that are mostly stored in a configuration
dict. Shared resources are also used later to setup an already saved reader.
"""

import os
import pickle

import yaml

from jack.io.embeddings import Embeddings
from jack.util.vocab import Vocab


class SharedResources:
    """Shared resources between modules.

    A class to provide and store generally shared resources, such as vocabularies,
    across the reader sub-modules.
    """

    def __init__(self, vocab: Vocab = None, config: dict = None, embeddings: Embeddings = None):
        """
        Several shared resources are initialised here, even if no arguments
        are passed when calling __init__.
        The instantiated objects will be filled by the InputModule.
        - self.config holds hyperparameter values and general configuration
            parameters.
        - self.vocab serves as default Vocabulary object.
        - self.answer_vocab is by default the same as self.vocab. However,
            this attribute can be changed by the InputModule, e.g. by setting
            sepvocab=True when calling the setup_from_data() of the InputModule.
        """
        self.config = config or dict()
        self.vocab = vocab
        self.embeddings = embeddings

    def store(self, path):
        """
        Saves all attributes of this object.

        Args:
            path: path to save shared resources
        """
        if not os.path.exists(path):
            os.mkdir(path)
        vocabs = [(k, v) for k, v in self.__dict__.items() if isinstance(v, Vocab)]
        with open(os.path.join(path, 'remainder'), 'wb') as f:
            remaining = {k: v for k, v in self.__dict__.items()
                         if not isinstance(v, Vocab) and not k == 'config' and not k == 'embeddings'}
            pickle.dump(remaining, f, pickle.HIGHEST_PROTOCOL)
        for k, v in vocabs:
            v.store(os.path.join(path, k))
        with open(os.path.join(path, 'config.yaml'), 'w') as f:
            yaml.dump(self.config, f)
        if self.embeddings is not None:
            self.embeddings.store(os.path.join(path, 'embeddings'))

    def load(self, path):
        """
        Loads this (potentially empty) resource from path (all object attributes).
        Args:
            path: path to shared resources
        """
        remainder_path = os.path.join(path, 'remainder')
        if os.path.exists(remainder_path):
            with open(remainder_path, 'rb') as f:
                self.__dict__.update(pickle.load(f))
        for f in os.listdir(path):
            if f == 'config.yaml':
                with open(os.path.join(path, f), 'r') as f:
                    self.config = yaml.load(f)
            elif f == 'embeddings':
                self.embeddings = Embeddings.from_dir(os.path.join(path, f))
            else:
                v = Vocab()
                v.load(os.path.join(path, f))
                self.__dict__[f] = v



from collections import defaultdict


def evaluate(reader, dataset, batch_size):
    answers = reader.process_dataset(dataset, batch_size, silent=False)

    confusion_matrix = defaultdict(lambda: defaultdict(int))

    for (q, a), pa in zip(dataset, answers):
        confusion_matrix[a[0].text][pa.text] += 1

    classes = sorted(confusion_matrix.keys())
    max_class = max(6, len(max(classes, key=len)))

    precision = dict()
    recall = dict()
    f1 = dict()

    confusion_matrix_string = ['\n', ' ' * max_class]
    for c in classes:
        confusion_matrix_string.append('\t')
        confusion_matrix_string.append(c)
        confusion_matrix_string.append(' ' * (max_class - len(c)))
    confusion_matrix_string.append('\n')
    for c1 in classes:
        confusion_matrix_string.append(c1)
        confusion_matrix_string.append(' ' * (max_class - len(c1)))
        for c2 in classes:
            confusion_matrix_string.append('\t')
            ct = str(confusion_matrix[c1][c2])
            confusion_matrix_string.append(ct)
            confusion_matrix_string.append(' ' * (max_class - len(ct)))
        confusion_matrix_string.append('\n')
        precision[c1] = confusion_matrix[c1][c1] / max(1.0, sum(p[c1] for p in confusion_matrix.values()))
        recall[c1] = confusion_matrix[c1][c1] / max(1.0, sum(confusion_matrix[c1].values()))
        f1[c1] = 2 * precision[c1] * recall[c1] / max(1.0, precision[c1] + recall[c1])

    accuracy = sum(confusion_matrix[c][c] for c in classes) / max(
        1.0, sum(sum(vs.values()) for vs in confusion_matrix.values()))

    return {
        'Accuracy': accuracy,
        'Precision': precision,
        'Recall': recall,
        'F1': f1,
        'Confusion Matrix': ''.join(confusion_matrix_string)
    }



import torch
import torch.nn as nn
import torch.nn.functional as F

from jack.core import SharedResources
from jack.core.tensorport import Ports
from jack.core.torch import PyTorchModelModule
from jack.readers.extractive_qa.shared import XQAPorts
from jack.util.torch import Highway
from jack.util.torch import embedding, misc, xqa
from jack.util.torch.rnn import BiLSTM


class FastQAPyTorchModelModule(PyTorchModelModule):
    """PyTorch implementation of FastQA."""

    # TODO: does not support multiparagraph yet
    _input_ports = [XQAPorts.emb_question, XQAPorts.question_length,
                    XQAPorts.emb_support, XQAPorts.support_length,
                    # char embedding inputs
                    XQAPorts.word_chars, XQAPorts.word_char_length,
                    XQAPorts.question_batch_words, XQAPorts.support_batch_words,
                    # feature input
                    XQAPorts.word_in_question,
                    # optional input, provided only during training
                    XQAPorts.correct_start, XQAPorts.answer2support_training,
                    XQAPorts.is_eval]

    _output_ports = [XQAPorts.start_scores, XQAPorts.end_scores,
                     XQAPorts.answer_span]
    _training_input_ports = [XQAPorts.start_scores, XQAPorts.end_scores,
                             XQAPorts.answer_span_target, XQAPorts.answer2support_training]
    _training_output_ports = [Ports.loss]

    @property
    def output_ports(self):
        return self._output_ports

    @property
    def input_ports(self):
        return self._input_ports

    @property
    def training_input_ports(self):
        return self._training_input_ports

    @property
    def training_output_ports(self):
        return self._training_output_ports

    def create_loss_module(self, shared_resources: SharedResources):
        return xqa.XQAMinCrossentropyLossModule()

    def create_prediction_module(self, shared_resources: SharedResources):
        return FastQAPyTorchModule(shared_resources)


class FastQAPyTorchModule(nn.Module):
    def __init__(self, shared_resources: SharedResources):
        super(FastQAPyTorchModule, self).__init__()
        self._shared_resources = shared_resources
        input_size = shared_resources.embeddings.shape[-1]
        size = shared_resources.config["repr_dim"]
        self._size = size
        self._with_char_embeddings = self._shared_resources.config.get("with_char_embeddings", False)

        # modules & parameters
        if self._with_char_embeddings:
            self._conv_char_embedding = embedding.ConvCharEmbeddingModule(
                len(shared_resources.char_vocab), size)
            self._embedding_projection = nn.Linear(size + input_size, size)
            self._embedding_highway = Highway(size, 1)
            self._v_wiq_w = nn.Parameter(torch.ones(1, 1, input_size + size))
            input_size = size
        else:
            self._v_wiq_w = nn.Parameter(torch.ones(1, 1, input_size))

        self._bilstm = BiLSTM(input_size + 2, size)
        self._answer_layer = FastQAAnswerModule(shared_resources)

        # [size, 2 * size]
        self._question_projection = nn.Parameter(torch.cat([torch.eye(size), torch.eye(size)], dim=1))
        self._support_projection = nn.Parameter(torch.cat([torch.eye(size), torch.eye(size)], dim=1))

    def forward(self, emb_question, question_length,
                emb_support, support_length,
                unique_word_chars, unique_word_char_length,
                question_words2unique, support_words2unique,
                word_in_question, correct_start, answer2support,
                is_eval):
        """fast_qa model

        Args:
            emb_question: [Q, L_q, N]
            question_length: [Q]
            emb_support: [Q, L_s, N]
            support_length: [Q]
            unique_word_chars
            unique_word_char_length
            question_words2unique
            support_words2unique
            word_in_question: [Q, L_s]
            correct_start: [A], only during training, i.e., is_eval=False
            answer2question: [A], only during training, i.e., is_eval=False
            is_eval: []

        Returns:
            start_scores [B, L_s, N], end_scores [B, L_s, N], span_prediction [B, 2]
        """
        # Some helpers
        float_tensor = torch.cuda.FloatTensor if emb_question.is_cuda else torch.FloatTensor
        long_tensor = torch.cuda.LongTensor if emb_question.is_cuda else torch.LongTensor
        batch_size = question_length.data.shape[0]
        max_question_length = question_length.max().data[0]
        support_mask = misc.mask_for_lengths(support_length)
        question_binary_mask = misc.mask_for_lengths(question_length, mask_right=False, value=1.0)

        if self._with_char_embeddings:
            # compute combined embeddings
            [char_emb_question, char_emb_support] = self._conv_char_embedding(
                unique_word_chars, unique_word_char_length, [question_words2unique, support_words2unique])

            emb_question = torch.cat([emb_question, char_emb_question], 2)
            emb_support = torch.cat([emb_support, char_emb_support], 2)

        # compute encoder features
        question_features = torch.autograd.Variable(torch.ones(batch_size, max_question_length, 2, out=float_tensor()))
        question_features = question_features.type_as(emb_question)

        v_wiqw = self._v_wiq_w
        # [B, L_q, L_s]
        wiq_w = torch.matmul(emb_question * v_wiqw, emb_support.transpose(1, 2))
        # [B, L_q, L_s]
        wiq_w = wiq_w + support_mask.unsqueeze(1)
        wiq_w = F.softmax(
            wiq_w.view(batch_size * max_question_length, -1), dim=1).view(batch_size, max_question_length, -1)
        # [B, L_s]
        wiq_w = torch.matmul(question_binary_mask.unsqueeze(1), wiq_w).squeeze(1)

        # [B, L , 2]
        support_features = torch.stack([word_in_question, wiq_w], dim=2)

        if self._with_char_embeddings:
            # highway layer to allow for interaction between concatenated embeddings
            emb_question = self._embedding_projection(emb_question)
            emb_support = self._embedding_projection(emb_support)
            emb_question = self._embedding_highway(emb_question)
            emb_support = self._embedding_highway(emb_support)

        # dropout
        dropout = self._shared_resources.config.get("dropout", 0.0)
        emb_question = F.dropout(emb_question, dropout, training=not is_eval)
        emb_support = F.dropout(emb_support, dropout, training=not is_eval)

        # extend embeddings with features
        emb_question_ext = torch.cat([emb_question, question_features], 2)
        emb_support_ext = torch.cat([emb_support, support_features], 2)

        # encode question and support
        # [B, L, 2 * size]
        encoded_question = self._bilstm(emb_question_ext)[0]
        encoded_support = self._bilstm(emb_support_ext)[0]

        # [B, L, size]
        encoded_support = F.tanh(F.linear(encoded_support, self._support_projection))
        encoded_question = F.tanh(F.linear(encoded_question, self._question_projection))

        start_scores, end_scores, predicted_start_pointer, predicted_end_pointer = \
            self._answer_layer(encoded_question, question_length, encoded_support, support_length,
                               correct_start, answer2support, is_eval)

        # no multi paragraph support yet
        doc_idx = torch.autograd.Variable(torch.zeros(predicted_start_pointer.data.shape[0], out=long_tensor()))
        span = torch.stack([doc_idx, predicted_start_pointer, predicted_end_pointer], 1)

        return start_scores, end_scores, span


class FastQAAnswerModule(nn.Module):
    def __init__(self, shared_resources: SharedResources):
        super(FastQAAnswerModule, self).__init__()
        self._size = shared_resources.config["repr_dim"]

        # modules
        self._linear_question_attention = nn.Linear(self._size, 1, bias=False)

        self._linear_q_start_q = nn.Linear(self._size, self._size)
        self._linear_q_start = nn.Linear(2 * self._size, self._size, bias=False)
        self._linear_start_scores = nn.Linear(self._size, 1, bias=False)

        self._linear_q_end_q = nn.Linear(self._size, self._size)
        self._linear_q_end = nn.Linear(3 * self._size, self._size, bias=False)
        self._linear_end_scores = nn.Linear(self._size, 1, bias=False)

    def forward(self, encoded_question, question_length, encoded_support, support_length,
                correct_start, answer2question, is_eval):
        # casting
        long_tensor = torch.cuda.LongTensor if encoded_question.is_cuda else torch.LongTensor
        answer2question = answer2question.type(long_tensor)

        # computing single time attention over question
        attention_scores = self._linear_question_attention(encoded_question)
        q_mask = misc.mask_for_lengths(question_length)
        attention_scores = attention_scores.squeeze(2) + q_mask
        question_attention_weights = F.softmax(attention_scores, dim=1)
        question_state = torch.matmul(question_attention_weights.unsqueeze(1),
                                      encoded_question).squeeze(1)

        # Prediction
        # start
        start_input = torch.cat([question_state.unsqueeze(1) * encoded_support, encoded_support], 2)

        q_start_state = self._linear_q_start(start_input) + self._linear_q_start_q(question_state).unsqueeze(1)
        start_scores = self._linear_start_scores(F.relu(q_start_state)).squeeze(2)

        support_mask = misc.mask_for_lengths(support_length)
        start_scores = start_scores + support_mask
        _, predicted_start_pointer = start_scores.max(1)

        def align(t):
            return torch.index_select(t, 0, answer2question)

        if is_eval:
            start_pointer = predicted_start_pointer
        else:
            # use correct start during training, because p(end|start) should be optimized
            start_pointer = correct_start.type(long_tensor)
            predicted_start_pointer = align(predicted_start_pointer)
            start_scores = align(start_scores)
            start_input = align(start_input)
            encoded_support = align(encoded_support)
            question_state = align(question_state)
            support_mask = align(support_mask)

        # end
        u_s = []
        for b, p in enumerate(start_pointer):
            u_s.append(encoded_support[b, p.data[0]])
        u_s = torch.stack(u_s)

        end_input = torch.cat([encoded_support * u_s.unsqueeze(1), start_input], 2)

        q_end_state = self._linear_q_end(end_input) + self._linear_q_end_q(question_state).unsqueeze(1)
        end_scores = self._linear_end_scores(F.relu(q_end_state)).squeeze(2)

        end_scores = end_scores + support_mask

        max_support = support_length.max().data[0]

        if is_eval:
            end_scores += misc.mask_for_lengths(start_pointer, max_support, mask_right=False)

        _, predicted_end_pointer = end_scores.max(1)

        return start_scores, end_scores, predicted_start_pointer, predicted_end_pointer



# -*- coding: utf-8 -*-
import logging

import numpy as np

logger = logging.getLogger(__name__)


def get_list_shape(xs):
    if isinstance(xs, int):
        shape = []
    else:
        shape = [len(xs)]
        for i, x in enumerate(xs):
            if isinstance(x, list) or isinstance(x, tuple):
                if len(shape) == 1:
                    shape.append(0)
                shape[1] = max(len(x), shape[1])
                for j, y in enumerate(x):
                    if isinstance(y, list):
                        if len(shape) == 2:
                            shape.append(0)
                        shape[2] = max(len(y), shape[2])
    return shape


def numpify(xs, pad=0, keys=None, dtypes=None):
    """Converts a dict or list of Python data into a dict of numpy arrays."""
    is_dict = isinstance(xs, dict)
    xs_np = {} if is_dict else [0] * len(xs)
    xs_iter = xs.items() if is_dict else enumerate(xs)

    for i, (key, x) in enumerate(xs_iter):
        try:
            if (keys is None or key in keys) and not isinstance(x, np.ndarray):
                shape = get_list_shape(x)
                dtype = dtypes[i] if dtypes is not None else np.int64
                x_np = np.full(shape, pad, dtype)

                nb_dims = len(shape)

                if nb_dims == 0:
                    x_np = x
                else:
                    def f(tensor, values):
                        t_shp = tensor.shape
                        if len(t_shp) > 1:
                            for _i, _values in enumerate(values):
                                f(tensor[_i], _values)
                        else:
                            tensor[0:len(values)] = [v for v in values]

                    f(x_np, x)

                xs_np[key] = x_np
            else:
                xs_np[key] = x
        except Exception as e:
            logger.error('Error numpifying value ' + str(x) + ' of key ' + str(key))
            raise e
    return xs_np



# -*- coding: utf-8 -*-

import logging

import numpy as np

logger = logging.getLogger(__name__)


def load_fasttext(stream, vocab=None):
    """Loads fastText file and merges it if optional vocabulary
    Args:
        stream (iterable): An opened filestream to the fastText file.
        vocab (dict=None): Word2idx dict of existing vocabulary.
    Returns:
        return_vocab (Vocabulary), lookup (matrix); Vocabulary contains the
                     word2idx and the matrix contains the embedded words.
    """
    logger.info('Loading fastText vectors ..')

    word2idx = {}
    vec_n, vec_size = map(int, stream.readline().split())
    lookup = np.empty([vocab.get_size() if vocab is not None else vec_n, vec_size], dtype=np.float)
    n = 0
    for line in stream:
        word, vec = line.rstrip().split(maxsplit=1)
        if vocab is None or word in vocab and word not in word2idx:
            word = word.decode('utf-8')
            idx = len(word2idx)
            word2idx[word] = idx
            # if idx > np.size(lookup, axis=0) - 1:
            #    lookup.resize([lookup.shape[0] + 500000, lookup.shape[1]])
            lookup[idx] = np.fromstring(vec, sep=' ')
        n += 1
    # lookup.resize([len(word2idx), dim])
    logger.info('Loading fastText vectors completed.')
    return word2idx, lookup



import copy
import logging

import tensorflow as tf

from jack.util.tf.interaction_layer import interaction_layer
from jack.util.tf.sequence_encoder import encoder

logger = logging.getLogger(__name__)


def _flatten(l):
    if isinstance(l, list):
        return [module for sl in l for module in _flatten(sl)]
    else:
        return [l]


def _unique_module_name(module, layer_depth):
    inp = module.get('input', '')
    inp_str = inp if isinstance(inp, str) else '_'.join(inp)
    name = '_'.join([str(layer_depth), inp_str, module['module']])
    return name


def modular_encoder(encoder_config, inputs, inputs_length, inputs_mapping, default_repr_dim, dropout, is_eval):
    outputs = dict(inputs)
    outputs_length = dict(inputs_length)
    outputs_mapping = dict(inputs_mapping)
    seen_layer = set()

    def encode_module(module):
        module_type = module['module']

        reuse = module['name'] in seen_layer
        seen_layer.add(module['name'])

        if module_type == 'repeat':
            reuse = module.get('reuse')
            for k in range(module['num']):
                prefix = module['name'] + '/' if reuse else '%s_%d/' % (module['name'], k)
                for j, inner_module in enumerate(module['encoder']):
                    # copy this configuration
                    inner_module = copy.deepcopy(inner_module)
                    if 'name' not in inner_module:
                        inner_module['name'] = _unique_module_name(inner_module, j)
                    inner_module['name'] = prefix + inner_module['name']
                    encode_module(inner_module)
            return

        try:
            key = module['input']
            out_key = module.get('output', key)
            if module['module'] in ['concat', 'add', 'mul', 'weighted_add', 'sub']:
                outputs_length[out_key] = outputs_length[key[0]]
                outputs_mapping[out_key] = outputs_mapping.get(key[0])
                if module['module'] == 'concat':
                    outputs[out_key] = tf.concat([outputs[k] for k in key], 2, name=module['name'])
                    return
                if module['module'] == 'add':
                    outputs[out_key] = tf.add_n([outputs[k] for k in key], name=module['name'])
                    return
                if module['module'] == 'sub':
                    outputs[out_key] = tf.subtract(outputs[key[0]], outputs[key[1]], name=module['name'])
                    return
                if module['module'] == 'mul':
                    o = outputs[key[0]]
                    for k in key[1:-1]:
                        o *= outputs[k]
                    outputs[out_key] = tf.multiply(o, outputs[key[-1]], name=module['name'])
                    return
                if module['module'] == 'weighted_add':
                    bias = module.get('bias', 0.0)
                    g = tf.layers.dense(tf.concat([outputs[k] for k in key], 2), outputs[key[0]].get_shape()[-1].value,
                                        tf.sigmoid, bias_initializer=tf.constant_initializer(bias))
                    outputs[out_key] = tf.identity(g * outputs[key[0]] + (1.0 - g) * outputs[key[0]],
                                                   name=module['name'])
                    return
            if 'repr_dim' not in module:
                module['repr_dim'] = default_repr_dim
            if 'dependent' in module:
                dep_key = module['dependent']
                outputs[out_key] = interaction_layer(
                    outputs[key], outputs_length[key],
                    outputs[dep_key], outputs_length[dep_key],
                    outputs_mapping.get(key), outputs_mapping.get(dep_key), reuse=reuse, **module)
            else:
                if module.get('dropout') is True:
                    # set dropout to default dropout
                    module['dropout'] = dropout
                outputs[out_key] = encoder(outputs[key], outputs_length[key], reuse=reuse, is_eval=is_eval, **module)
            outputs_length[out_key] = outputs_length[key]
            outputs_mapping[out_key] = outputs_mapping.get(key)
        except Exception as e:
            logger.error('Creating module %s failed.', module['name'])
            raise e

    encoder_config = _flatten(encoder_config)
    # don't change original config but copy here
    encoder_config = copy.deepcopy(encoder_config)

    for i, module in enumerate(encoder_config):
        if 'name' not in module:
            module['name'] = _unique_module_name(module, i)
        encode_module(module)

    return outputs, outputs_length, outputs_mapping



# -*- coding: utf-8 -*-

import operator
import pickle
import sys
from collections import OrderedDict


class Vocab:
    """
    Vocab objects for use in jack pipelines.
    """
    DEFAULT_UNK = "<UNK>"

    def __init__(self, unk=DEFAULT_UNK, vocab: dict = None):
        """
        Creates Vocab object.

        Args:
            `unk`: symbol for unknown term (default: "<UNK>").
              If set to `None`, and `None` is not included as symbol while unfrozen,
              it will return `None` upon calling `get_id(None)` when frozen.
            `vocab`: init from vocab dict (sym -> id)
        """
        self._unk = unk
        if vocab is not None:
            self._sym2id = dict(vocab)
            self._id2sym = {v: k for k, v in vocab.items()}
            if unk is not None and unk not in self._sym2id:
                self._sym2id[unk] = len(self._sym2id)
                self._id2sym[len(self._id2sym)] = unk
            self._sym2freqs = {w: 0 for w in self._sym2id}
            self._frozen = True
        else:
            self._sym2id = {}
            # with pos and neg indices
            self._id2sym = {}
            self._sym2freqs = OrderedDict()
            if unk is not None:
                self._sym2id[unk] = 0
                # with pos and neg indices
                self._id2sym[0] = unk
                self._sym2freqs[unk] = 0
            self._frozen = False

    def freeze(self):
        """Freeze current Vocab object (set `self._frozen` to True)."""
        # if any pretrained have been encountered
        self._frozen = True

    def unfreeze(self):
        """Unfreeze current Vocab object (set `self.frozen` to False)."""
        self._frozen = False

    def get_id(self, sym):
        """Get id of symbol. Counts frequency if not frozen."""
        if not self._frozen:
            if sym not in self._sym2id:
                self._sym2id[sym] = len(self._sym2id)
                self._id2sym[len(self._id2sym)] = sym
                self._sym2freqs[sym] = 1
            else:
                self._sym2freqs[sym] += 1
        return self._sym2id.get(sym, self._sym2id.get(self._unk))

    def get_sym(self, id):
        """returns symbol for a given id (consistent with the `self.frozen` state), and None if not found."""
        return self._id2sym.get(id)

    def __call__(self, *args, **kwargs):
        """
        calls the `get_id` function for the provided symbol(s), which adds symbols to the Vocab if needed and allowed,
        and returns their id(s).

        Args:
            *args: a single symbol, a list of symbols, or multiple symbols
        """
        symbols = args
        if len(args) == 1:
            if isinstance(args[0], list):
                symbols = args[0]
            else:
                return self.get_id(args[0])
        return [self.get_id(sym) for sym in symbols]

    def __len__(self):
        """returns number of unique symbols (including the unknown symbol)"""
        return len(self._id2sym)

    def __contains__(self, sym):
        """checks if `sym` already in the Vocab object"""
        return sym in self._sym2id

    @property
    def frozen(self):
        return self._frozen

    @property
    def unk(self):
        return self._unk

    def prune(self, min_freq=5, max_size=sys.maxsize):
        """returns new Vocab object, pruned based on minimum symbol frequency"""
        pruned_vocab = Vocab(unk=self._unk)
        cnt = 0
        for sym, freq in sorted(self._sym2freqs.items(), key=operator.itemgetter(1), reverse=True):
            # for sym in self.sym2freqs:
            # freq = self.sym2freqs[sym]
            cnt += 1
            if freq >= min_freq and cnt < max_size:
                pruned_vocab(sym)
                pruned_vocab._sym2freqs[sym] = freq
        if self._frozen:
            # if original Vocab was frozen, freeze new one
            pruned_vocab.freeze()

        return pruned_vocab

    def store(self, path: str):
        with open(path, "wb") as f:
            pickle.dump(self.__dict__, f)

    def load(self, path: str):
        with open(path, "rb") as f:
            data = pickle.load(f)
        # backwards compability
        new_data = {}
        for k in data:
            if k not in self.__dict__:
                new_k = '_' + k
                if new_k in self.__dict__:
                    new_data[new_k] = data[k]
            else:
                new_data[k] = data[k]
        self.__dict__ = new_data



"""
This file contains reusable modules for extractive QA models and ports
"""
from typing import NamedTuple

from jack.core import *
from jack.readers.extractive_qa.util import prepare_data
from jack.util import preprocessing
from jack.util.map import numpify
from jack.util.preprocessing import sort_by_tfidf

logger = logging.getLogger(__name__)


class XQAPorts:
    # When feeding embeddings directly
    emb_question = Ports.Input.emb_question
    question_length = Ports.Input.question_length
    emb_support = Ports.Input.emb_support
    support_length = Ports.Input.support_length
    support2question = Ports.Input.support2question

    # but also ids, for char-based embeddings
    word_chars = Ports.Input.word_chars
    word_char_length = Ports.Input.word_char_length
    question_batch_words = Ports.Input.question_batch_words
    support_batch_words = Ports.Input.support_batch_words

    is_eval = Ports.is_eval

    # This feature is model specific and thus, not part of the conventional Ports
    word_in_question = TensorPort(np.float32, [None, None], "word_in_question",
                                  "Represents a 1/0 feature for all context tokens denoting"
                                  " whether it is part of the question or not",
                                  "[Q, support_length]")

    correct_start = TensorPortWithDefault(np.array([0], np.int32), [None], "correct_start",
                                          "Represents the correct start of the span which is given to the"
                                          "model during training for use to predicting end.",
                                          "[A]")

    # output ports
    start_scores = Ports.Prediction.start_scores
    end_scores = Ports.Prediction.end_scores
    answer_span = Ports.Prediction.answer_span
    token_offsets = TensorPort(np.int32, [None, None], "token_offsets",
                               "Character index of tokens in support.",
                               "[S, support_length]")
    selected_support = TensorPort(np.int32, [None], "selected_support",
                                  "Selected support based on TF IDF with question", "[num_support]")

    # ports used during training
    answer2support_training = Ports.Input.answer2support
    answer_span_target = Ports.Target.answer_span


XQAAnnotation = NamedTuple('XQAAnnotation', [
    ('question_tokens', List[str]),
    ('question_ids', List[int]),
    ('question_length', int),
    ('support_tokens', List[List[str]]),
    ('support_ids', List[List[int]]),
    ('support_length', List[int]),
    ('word_in_question', List[List[float]]),
    ('token_offsets', List[List[int]]),
    ('answer_spans', Optional[List[List[Tuple[int, int]]]]),
    ('selected_supports', Optional[List[int]]),
])


class XQAInputModule(OnlineInputModule[XQAAnnotation]):
    _output_ports = [XQAPorts.emb_question, XQAPorts.question_length,
                     XQAPorts.emb_support, XQAPorts.support_length,
                     XQAPorts.support2question,
                     # char
                     XQAPorts.word_chars, XQAPorts.word_char_length,
                     XQAPorts.question_batch_words, XQAPorts.support_batch_words,
                     # features
                     XQAPorts.word_in_question,
                     # optional, only during training
                     XQAPorts.correct_start, XQAPorts.answer2support_training,
                     XQAPorts.is_eval,
                     # for output module
                     XQAPorts.token_offsets, XQAPorts.selected_support]
    _training_ports = [XQAPorts.answer_span_target, XQAPorts.answer2support_training]

    def setup_from_data(self, data: Iterable[Tuple[QASetting, List[Answer]]]):
        # create character vocab + word lengths + char ids per word
        if not self.shared_resources.vocab.frozen:
            preprocessing.fill_vocab((q for q, _ in data), self.shared_resources.vocab,
                                     self.shared_resources.config.get("lowercase", False))
            self.shared_resources.vocab.freeze()
        self.shared_resources.char_vocab = preprocessing.char_vocab_from_vocab(self.shared_resources.vocab)

    def setup(self):
        self._rng = random.Random(1)
        self.vocab = self.shared_resources.vocab
        self.config = self.shared_resources.config
        self.embeddings = self.shared_resources.embeddings
        self.__default_vec = np.zeros([self.embeddings.shape[-1]])
        self.char_vocab = self.shared_resources.char_vocab

    @property
    def output_ports(self) -> List[TensorPort]:
        return self._output_ports

    @property
    def training_ports(self) -> List[TensorPort]:
        return self._training_ports

    def preprocess(self, questions: List[QASetting],
                   answers: Optional[List[List[Answer]]] = None,
                   is_eval: bool = False) -> List[XQAAnnotation]:

        if answers is None:
            answers = [None] * len(questions)
        preprocessed = []
        if len(questions) > 1000:
            bar = progressbar.ProgressBar(
                max_value=len(questions),
                widgets=[' [', progressbar.Timer(), '] ', progressbar.Bar(), ' (', progressbar.ETA(), ') '])
            for q, a in bar(zip(questions, answers)):
                preprocessed.append(self.preprocess_instance(q, a))
        else:
            for q, a in zip(questions, answers):
                preprocessed.append(self.preprocess_instance(q, a))

        return preprocessed

    def preprocess_instance(self, question: QASetting, answers: Optional[List[Answer]] = None) -> XQAAnnotation:
        has_answers = answers is not None

        q_tokenized, q_ids, _, q_length, s_tokenized, s_ids, _, s_length, \
        word_in_question, token_offsets, answer_spans = prepare_data(
            question, answers, self.vocab, self.config.get("lowercase", False),
            with_answers=has_answers, max_support_length=self.config.get("max_support_length", None))

        max_num_support = self.config.get("max_num_support", len(question.support))  # take all per default

        # take max supports by TF-IDF (we subsample to max_num_support in create batch)
        # following https://arxiv.org/pdf/1710.10723.pdf
        if len(question.support) > 1:
            scores = sort_by_tfidf(' '.join(q_tokenized), [' '.join(s) for s in s_tokenized])
            selected_supports = [s_idx for s_idx, _ in scores[:max_num_support]]
            s_tokenized = [s_tokenized[s_idx] for s_idx in selected_supports]
            s_ids = [s_ids[s_idx] for s_idx in selected_supports]
            s_length = [s_length[s_idx] for s_idx in selected_supports]
            word_in_question = [word_in_question[s_idx] for s_idx in selected_supports]
            token_offsets = [token_offsets[s_idx] for s_idx in selected_supports]
            answer_spans = [answer_spans[s_idx] for s_idx in selected_supports]
        else:
            selected_supports = list(range(len(question.support)))

        return XQAAnnotation(
            question_tokens=q_tokenized,
            question_ids=q_ids,
            question_length=q_length,
            support_tokens=s_tokenized,
            support_ids=s_ids,
            support_length=s_length,
            word_in_question=word_in_question,
            token_offsets=token_offsets,
            answer_spans=answer_spans if has_answers else None,
            selected_supports=selected_supports,
        )

    def create_batch(self, annotations: List[XQAAnnotation], is_eval: bool, with_answers: bool) \
            -> Mapping[TensorPort, np.ndarray]:

        q_tokenized = [a.question_tokens for a in annotations]
        question_lengths = [a.question_length for a in annotations]

        max_training_support = self.config.get('max_training_support', 2)
        s_tokenized = []
        support_lengths = []
        wiq = []
        offsets = []
        support2question = []
        support_ids = []
        # aligns with support2question, used in output module to get correct index to original set of supports
        selected_support = []
        all_spans = []
        for i, a in enumerate(annotations):
            all_spans.append([])
            if len(a.support_tokens) > max_training_support > 0 and not is_eval:
                # sample only 2 paragraphs and take first with double probability (the best) to speed
                # things up. Following https://arxiv.org/pdf/1710.10723.pdf
                is_done = False
                any_answer = any(a.answer_spans)
                # sample until there is at least one possible answer (if any)
                while not is_done:
                    selected = self._rng.sample(range(0, len(a.support_tokens) + 1), max_training_support + 1)
                    if 0 in selected and 1 in selected:
                        selected = [s - 1 for s in selected if s > 0]
                    else:
                        selected = [max(0, s - 1) for s in selected[:max_training_support]]
                    is_done = not any_answer or any(a.answer_spans[s] for s in selected)
            else:
                selected = set(range(len(a.support_tokens)))
            for s in selected:
                s_tokenized.append(a.support_tokens[s])
                support_lengths.append(a.support_length[s])
                wiq.append(a.word_in_question[s])
                offsets.append(a.token_offsets[s])
                selected_support.append(a.selected_supports[s])
                support_ids.append(a.support_ids[s])
                support2question.append(i)
                if with_answers:
                    all_spans[-1].append(a.answer_spans[s])

        word_chars, word_lengths, batch_word_ids, batch_vocab, batch_rev_vocab = \
            preprocessing.unique_words_with_chars(q_tokenized + s_tokenized, self.char_vocab)

        emb_support = np.zeros([len(support_lengths), max(support_lengths), self.embeddings.shape[-1]])
        emb_question = np.zeros([len(question_lengths), max(question_lengths), self.embeddings.shape[-1]])

        for i, a in enumerate(annotations):
            for j, t in enumerate(a.question_tokens):
                emb_question[i, j] = self.embeddings.get(t, self.__default_vec)
        for k, s_ids in enumerate(s_tokenized):
            for j, t in enumerate(s_ids):
                emb_support[k, j] = self.embeddings.get(t, self.__default_vec)

        output = {
            XQAPorts.word_chars: word_chars,
            XQAPorts.word_char_length: word_lengths,
            XQAPorts.question_batch_words: batch_word_ids[:len(q_tokenized)],
            XQAPorts.support_batch_words: batch_word_ids[len(q_tokenized):],
            XQAPorts.emb_support: emb_support,
            XQAPorts.support_length: support_lengths,
            XQAPorts.emb_question: emb_question,
            XQAPorts.question_length: question_lengths,
            XQAPorts.word_in_question: wiq,
            XQAPorts.support2question: support2question,
            XQAPorts.is_eval: is_eval,
            XQAPorts.token_offsets: offsets,
            XQAPorts.selected_support: selected_support,
            '__vocab': batch_vocab,
            '__rev_vocab': batch_rev_vocab,
        }

        if with_answers:
            spans = [s for a in all_spans for spans_per_support in a for s in spans_per_support]
            span2support = []
            support_idx = 0
            for a in all_spans:
                for spans_per_support in a:
                    span2support.extend([support_idx] * len(spans_per_support))
                    support_idx += 1
            output.update({
                XQAPorts.answer_span_target: [span for span in spans] if spans else np.zeros([0, 2], np.int32),
                XQAPorts.correct_start: [] if is_eval else [span[0] for span in spans],
                XQAPorts.answer2support_training: span2support,
            })

        # we can only numpify in here, because bucketing is not possible prior
        batch = numpify(output, keys=[XQAPorts.word_chars,
                                      XQAPorts.question_batch_words, XQAPorts.support_batch_words,
                                      XQAPorts.word_in_question, XQAPorts.token_offsets])
        return batch


def _np_softmax(x):
    """Compute softmax values for each sets of scores in x."""
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum(axis=0)


def get_answer_and_span(question, doc_idx, start, end, token_offsets, selected_support):
    doc_idx = selected_support[doc_idx]
    char_start = token_offsets[start]
    char_end = token_offsets[end + 1] if end < len(token_offsets) - 1 else len(question.support[doc_idx])
    answer = question.support[doc_idx][char_start: char_end]
    answer = answer.rstrip()
    char_end = char_start + len(answer)
    return answer, doc_idx, (char_start, char_end)


class XQAOutputModule(OutputModule):
    @property
    def input_ports(self) -> List[TensorPort]:
        return [XQAPorts.answer_span, XQAPorts.token_offsets,
                XQAPorts.selected_support, XQAPorts.support2question,
                XQAPorts.start_scores, XQAPorts.end_scores]

    def __call__(self, questions, tensors: Mapping[TensorPort, np.ndarray]) -> Sequence[Sequence[Answer]]:
        """Produces top-k answers for each question."""
        tensors = TensorPortTensors(tensors)
        topk = tensors.answer_span.shape[0] // len(questions)
        all_answers = []
        for n, q in enumerate(questions):
            answers = []
            doc_idx_map = [i for i, q_id in enumerate(tensors.support2question) if q_id == n]
            for j in range(topk):
                i = n * topk + j
                doc_idx, start, end = tensors.answer_span[i]
                score = (_np_softmax(tensors.start_scores[doc_idx_map[doc_idx]])[start] *
                         _np_softmax(tensors.end_scores[doc_idx_map[doc_idx]])[end])
                answer, doc_idx, span = get_answer_and_span(
                    q, doc_idx, start, end, tensors.token_offsets[doc_idx_map[doc_idx]],
                    [i for q_id, i in zip(tensors.support2question, tensors.selected_support) if q_id == n])
                answers.append(Answer(answer, span=span, doc_idx=doc_idx, score=score))
            all_answers.append(answers)

        return all_answers




# -*- coding: utf-8 -*-

import logging
import math
import os
import random
import shutil

import tensorflow as tf

from jack import readers
from jack.core.tensorflow import TFReader
from jack.eval import evaluate_reader, pretty_print_results
from jack.util.hooks import LossHook, ExamplesPerSecHook, ETAHook

logger = logging.getLogger(__name__)


def train(reader, train_data, test_data, dev_data, configuration: dict, debug=False):
    if isinstance(reader, TFReader):
        train_tensorflow(reader, train_data, test_data, dev_data, configuration, debug)
    else:
        train_pytorch(reader, train_data, test_data, dev_data, configuration, debug)


def train_tensorflow(reader, train_data, test_data, dev_data, configuration: dict, debug=False):
    import tensorflow as tf
    seed = configuration.get('seed', 0)

    # make everything deterministic
    random.seed(seed)
    tf.set_random_seed(seed)

    clip_value = configuration.get('clip_value')
    batch_size = configuration.get('batch_size')
    dev_batch_size = configuration.get('dev_batch_size') or batch_size
    epochs = configuration.get('epochs')
    l2 = configuration.get('l2')
    optimizer = configuration.get('optimizer')
    learning_rate = configuration.get('learning_rate')
    min_learning_rate = configuration.get('min_learning_rate')
    learning_rate_decay = configuration.get('learning_rate_decay')
    log_interval = configuration.get('log_interval')
    validation_interval = configuration.get('validation_interval')
    tensorboard_folder = configuration.get('tensorboard_folder')
    reader_type = configuration.get('reader')
    save_dir = configuration.get('save_dir')
    write_metrics_to = configuration.get('write_metrics_to')

    if clip_value != 0.0:
        clip_value = - abs(clip_value), abs(clip_value)

    learning_rate = tf.get_variable("learning_rate", initializer=learning_rate, dtype=tf.float32, trainable=False)
    lr_decay_op = learning_rate.assign(tf.maximum(learning_rate_decay * learning_rate, min_learning_rate))

    name_to_optimizer = {
        'gd': tf.train.GradientDescentOptimizer,
        'adam': tf.train.AdamOptimizer,
        'adagrad': tf.train.AdagradOptimizer,
        'adadelta': tf.train.AdadeltaOptimizer,
        'rmsprop': tf.train.RMSPropOptimizer
    }

    if optimizer not in name_to_optimizer:
        raise ValueError('Unknown optimizer: {}'.format(optimizer))

    tf_optimizer_class = name_to_optimizer[optimizer]
    tf_optimizer = tf_optimizer_class(learning_rate=learning_rate)

    sw = None
    if tensorboard_folder is not None:
        if os.path.exists(tensorboard_folder):
            shutil.rmtree(tensorboard_folder)
        sw = tf.summary.FileWriter(tensorboard_folder)

    # Hooks
    iter_interval = 1 if debug else log_interval
    hooks = [LossHook(reader, iter_interval, summary_writer=sw),
             ETAHook(reader, iter_interval, int(math.ceil(len(train_data) / batch_size)), epochs),
             ExamplesPerSecHook(reader, batch_size, iter_interval, sw)]

    preferred_metric, best_metric = readers.eval_hooks[reader_type].preferred_metric_and_initial_score()

    def side_effect(metrics, prev_metric):
        """Returns: a state (in this case a metric) that is used as input for the next call"""
        m = metrics[preferred_metric]
        if prev_metric is not None and m < prev_metric:
            reader.session.run(lr_decay_op)
            logger.info("Decayed learning rate to: %.5f" % reader.session.run(learning_rate))
        elif m > best_metric[0] and save_dir is not None:
            best_metric[0] = m
            if prev_metric is None:  # store whole reader_type only at beginning of training
                reader.store(save_dir)
            else:
                reader.model_module.store(os.path.join(save_dir, "model_module"))
            logger.info("Saving reader_type to: %s" % save_dir)
        return m

    # this is the standard hook for the reader_type
    hooks.append(readers.eval_hooks[reader_type](
        reader, dev_data, dev_batch_size, summary_writer=sw, side_effect=side_effect,
        iter_interval=validation_interval,
        epoch_interval=(1 if validation_interval is None else None),
        write_metrics_to=write_metrics_to))

    # Train
    reader.train(tf_optimizer, train_data, batch_size, max_epochs=epochs, hooks=hooks,
                 l2=l2, clip=clip_value, clip_op=tf.clip_by_value, summary_writer=sw)

    # Test final reader_type
    if dev_data is not None and save_dir is not None:
        reader.load(save_dir)
        result_dict = evaluate_reader(reader, dev_data, batch_size)

        logger.info("############### Results on the Dev Set##############")
        pretty_print_results(result_dict)

    if test_data is not None and save_dir is not None:
        reader.load(save_dir)
        result_dict = evaluate_reader(reader, test_data, batch_size)

        logger.info("############### Results on the Test Set##############")
        pretty_print_results(result_dict)


def train_pytorch(reader, train_data, test_data, dev_data, configuration: dict, debug=False):
    import torch
    seed = configuration.get('seed')

    # make everything deterministic
    random.seed(seed)
    torch.manual_seed(seed)

    clip_value = configuration.get('clip_value')
    batch_size = configuration.get('batch_size')
    epochs = configuration.get('epochs')
    l2 = configuration.get('l2')
    optimizer = configuration.get('optimizer')
    learning_rate = configuration.get('learning_rate')
    learning_rate_decay = configuration.get('learning_rate_decay')
    log_interval = configuration.get('log_interval')
    validation_interval = configuration.get('validation_interval')
    tensorboard_folder = configuration.get('tensorboard_folder')
    model = configuration.get('reader')
    save_dir = configuration.get('save_dir')
    write_metrics_to = configuration.get('write_metrics_to')

    # need setup here already :(
    reader.setup_from_data(train_data, is_training=True)

    if clip_value != 0.0:
        clip_value = - abs(clip_value), abs(clip_value)

    name_to_optimizer = {
        'gd': torch.optim.SGD,
        'adam': torch.optim.Adam,
        'adagrad': torch.optim.Adagrad,
        'adadelta': torch.optim.Adadelta
    }

    if optimizer not in name_to_optimizer:
        raise ValueError('Unknown optimizer: {}'.format(optimizer))

    torch_optimizer_class = name_to_optimizer[optimizer]
    params = list(reader.model_module.prediction_module.parameters())
    params.extend(reader.model_module.loss_module.parameters())

    torch_optimizer = torch_optimizer_class(params, lr=learning_rate)

    sw = None
    if tensorboard_folder is not None:
        if os.path.exists(tensorboard_folder):
            shutil.rmtree(tensorboard_folder)
        sw = tf.summary.FileWriter(tensorboard_folder)

    # Hooks
    iter_interval = 1 if debug else log_interval
    hooks = [LossHook(reader, iter_interval, summary_writer=sw),
             ExamplesPerSecHook(reader, batch_size, iter_interval, sw)]

    preferred_metric, best_metric = readers.eval_hooks[model].preferred_metric_and_initial_score()

    def side_effect(metrics, prev_metric):
        """Returns: a state (in this case a metric) that is used as input for the next call"""
        m = metrics[preferred_metric]
        if prev_metric is not None and m < prev_metric:
            for param_group in torch_optimizer.param_groups:
                param_group['lr'] *= learning_rate_decay
                logger.info("Decayed learning rate to: %.5f" % param_group['lr'])
        elif m > best_metric[0] and save_dir is not None:
            best_metric[0] = m
            if prev_metric is None:  # store whole model only at beginning of training
                reader.store(save_dir)
            else:
                reader.model_module.store(os.path.join(save_dir, "model_module"))
            logger.info("Saving model to: %s" % save_dir)
        return m

    # this is the standard hook for the model
    hooks.append(readers.eval_hooks[model](
        reader, dev_data, batch_size, summary_writer=sw, side_effect=side_effect,
        iter_interval=validation_interval,
        epoch_interval=(1 if validation_interval is None else None),
        write_metrics_to=write_metrics_to))

    # Train
    reader.train(torch_optimizer, train_data, batch_size, max_epochs=epochs, hooks=hooks,
                 l2=l2, clip=clip_value)

    # Test final model
    if dev_data is not None and save_dir is not None:
        reader.load(save_dir)
        result_dict = evaluate_reader(reader, dev_data, batch_size)

        logger.info("############### Results on the Dev Set##############")
        pretty_print_results(result_dict)

    if test_data is not None and save_dir is not None:
        reader.load(save_dir)
        result_dict = evaluate_reader(reader, test_data, batch_size)

        logger.info("############### Results on the Test Set##############")
        pretty_print_results(result_dict)



import json
import io


def create_snippet(file_path, first_n=5):
    with open(file_path, 'r') as f:
        return [next(f) for _ in range(first_n)]


def create_jtr_snippet(file_path):
    return convert_simplequestions(file_path, first_n=5)


def convert_simplequestions(file_path, first_n=None):
    instances = []
    f = io.open(file_path, "r")
    i = 0
    for l in f:
        i += 1
        if first_n and i > first_n:
            break
        subj, rel, obj, qu = l.strip().split("\t")

        support = [" ".join([subj, rel])]
        qdict = {
            'question': qu,
            'answers': [obj]
        }
        qset_dict = {
            'support': [{'text': supp} for supp in support],
            'questions': [qdict]
        }
        instances.append(qset_dict)

    corpus_dict = {
        'meta': "simpleQuestions.json",
        'instances': instances
    }

    f.close()

    return corpus_dict


def main():
    # some tests:
    # raw_data = load_cbt_file(path=None, part='valid', mode='NE')
    # instances = split_cbt(raw_data)
    # = parse_cbt_example(instances[0])

    import sys
    if len(sys.argv) == 3:
        # corpus = create_jtr_snippet(sys.argv[1])
        # out = create_snippet(sys.argv[1])
        # with open(sys.argv[2], 'w') as outfile:
        #     outfile.writelines(out)
        corpus = convert_simplequestions(sys.argv[1])
        with open(sys.argv[2], 'w') as outfile:
            json.dump(corpus, outfile, indent=2)
    else:
        print("Usage: python3 simpleQuestions2jtr.py path/to/simpleQuestions save/to/simpleQuestions.jack.json")


if __name__ == "__main__":
    main()



from jack.core.input_module import *
from jack.core.model_module import *
from jack.core.output_module import *
from jack.core.reader import *
from jack.core.tensorport import *
from jack.core.shared_resources import *



import tensorflow as tf


def segment_softmax(scores, segment_ids):
    """Given scores and a partition, converts scores to probs by performing
    softmax over all rows within a partition."""

    # Subtract max
    num_segments = tf.reduce_max(segment_ids) + 1
    if len(scores.get_shape()) == 2:
        max_per_partition = tf.unsorted_segment_max(tf.reduce_max(scores, axis=1), segment_ids, num_segments)
        scores -= tf.expand_dims(tf.gather(max_per_partition, segment_ids), axis=1)
    else:
        max_per_partition = tf.unsorted_segment_max(scores, segment_ids, num_segments)
        scores -= tf.gather(max_per_partition, segment_ids)

    # Compute probs
    scores_exp = tf.exp(scores)
    if len(scores.get_shape()) == 2:
        scores_exp_sum_per_partition = tf.unsorted_segment_sum(tf.reduce_sum(scores_exp, axis=1), segment_ids,
                                                               num_segments)
        probs = scores_exp / tf.expand_dims(tf.gather(scores_exp_sum_per_partition, segment_ids), axis=1)
    else:
        scores_exp_sum_per_partition = tf.unsorted_segment_sum(scores_exp, segment_ids, num_segments)
        probs = scores_exp / tf.gather(scores_exp_sum_per_partition, segment_ids)

    return probs


def segment_argmax(input, segment_ids):
    """Computes row and col indices Tensors of the segment max in the 2D input."""

    with tf.name_scope("segment_argmax"):
        num_partitions = tf.reduce_max(segment_ids) + 1
        is_max = segment_is_max(input, segment_ids)

        # The current is_max could still contain multiple True entries per
        # partition. As long as they are in the same row, that is not a problem.
        # However, we do need to remove duplicate Trues in the same partition
        # in multiple rows.
        # For that, we'll multiply is_max with the row indices + 1 and perform
        # segment_is_max() again.

        rows = tf.shape(input)[0]
        cols = tf.shape(input)[1]
        row_indices = tf.tile(tf.expand_dims(tf.range(rows), 1), [1, cols])
        is_max = segment_is_max(tf.cast(is_max, tf.int32) * (row_indices + 1), segment_ids)

        # Get selected rows and columns
        row_selected = tf.reduce_any(is_max, axis=1)
        row_indices = tf.squeeze(tf.where(row_selected))
        rows_selected = tf.reduce_sum(tf.cast(row_selected, tf.int64))

        # Assert rows_selected is correct & ensure row_indices is always 1D
        with tf.control_dependencies([tf.assert_equal(rows_selected, num_partitions)]):
            row_indices = tf.reshape(row_indices, [-1])

        selected_rows_is_max = tf.gather(is_max, row_indices)
        col_indices = tf.argmax(tf.cast(selected_rows_is_max, tf.int64), axis=1)

        # Pack indices
        return row_indices, col_indices


def segment_is_max(inputs, segment_ids):
    num_segments = tf.reduce_max(segment_ids) + 1
    if len(inputs.get_shape()) > 1:
        inputs_max = tf.reduce_max(inputs, reduction_indices=list(range(1, len(inputs.get_shape()))))
    else:
        inputs_max = inputs
    max_per_partition = tf.unsorted_segment_max(inputs_max, segment_ids, num_segments)
    return tf.equal(inputs, tf.gather(max_per_partition, segment_ids))


def segment_sample_select(probs, segment_ids):
    num_segments = tf.reduce_max(segment_ids) + 1
    sampled = tf.random_uniform([num_segments])

    def scan_fn(acc, x):
        p, i = x[0], x[1]
        prev_v = tf.gather(acc[0], i)
        new_probs = acc[0] + tf.one_hot(i, num_segments, p)
        select = tf.logical_and(tf.less(prev_v, 0.0), tf.greater_equal(prev_v + p, 0.0))
        return new_probs, select

    _, selection = tf.scan(scan_fn, (probs, segment_ids), initializer=(-sampled, False))

    return selection


def segment_top_k(input, segment_ids, k):
    """Computes top k elements for segments in 2D input.

    segment_idx needs to be sorted.

    Returns:
        [num_segments, k]- tensors for rows, columns, scores of best k results in each segment
    """

    with tf.name_scope("segment_top_k"):
        all_top_k_scores, all_top_k_indices = tf.nn.top_k(input, k)
        rows = tf.tile(tf.expand_dims(tf.range(tf.shape(input)[0], dtype=tf.int32), 1), [1, k])

        result_rows = tf.zeros([k], tf.int32)
        result_columns = tf.zeros([k], tf.int32)
        result_scores = tf.zeros([k], tf.float32)

        def replace(old, new):
            return tf.concat([old[:-1], tf.expand_dims(new, 0)], 0)

        def scan_fn(acc, x):
            result_row, result_column, result_score, last_index = acc

            row_indices = x[0]
            segment_idx = x[1]
            top_k_scores = x[2]
            col_indices = x[3]

            def merge():
                new_result_row = tf.concat([result_row, row_indices], 0)
                new_result_column = tf.concat([result_column, col_indices], 0)
                new_result_score = tf.concat([result_score, top_k_scores], 0)
                new_result_score, new_top_k_indices = tf.nn.top_k(new_result_score, k)
                new_result_row = tf.gather(new_result_row, new_top_k_indices[:k])
                new_result_column = tf.gather(new_result_column, new_top_k_indices[:k])

                return new_result_row, new_result_column, new_result_score, segment_idx

            return tf.cond(segment_idx > last_index,
                           lambda: (row_indices, col_indices, top_k_scores, segment_idx),
                           merge)

        last_index = -1
        result_rows, result_columns, result_scores, _ = tf.scan(
            scan_fn, (rows, segment_ids, all_top_k_scores, all_top_k_indices),
            initializer=(result_rows, result_columns, result_scores, last_index))

        to_gather = tf.squeeze(tf.where((segment_ids[1:] - segment_ids[:-1]) > 0))
        to_gather = tf.concat([to_gather, tf.shape(segment_ids, out_type=tf.int64) - 1], 0)

        return tf.gather(result_rows, to_gather), tf.gather(result_columns, to_gather), tf.gather(result_scores,
                                                                                                  to_gather)



# -*- coding: utf-8 -*-

import numpy as np
import tensorflow as tf


def birnn_with_projection(size, fused_rnn, inputs, length, share_rnn=False, projection_scope=None):
    projection_initializer = tf.constant_initializer(np.concatenate([np.eye(size), np.eye(size)]))
    with tf.variable_scope("RNN", reuse=share_rnn):
        encoded = fused_birnn(fused_rnn, inputs, sequence_length=length, dtype=tf.float32, time_major=False)[0]
        encoded = tf.concat(encoded, 2)

    projected = tf.layers.dense(encoded, size, kernel_initializer=projection_initializer, name=projection_scope)
    return projected


def fused_rnn_backward(fused_rnn, inputs, sequence_length, initial_state=None, dtype=None, scope=None, time_major=True):
    if not time_major:
        inputs = tf.transpose(inputs, [1, 0, 2])
    # assumes that time dim is 0 and batch is 1
    rev_inputs = tf.reverse_sequence(inputs, sequence_length, 0, 1)
    rev_outputs, last_state = fused_rnn(rev_inputs, sequence_length=sequence_length, initial_state=initial_state,
                                        dtype=dtype, scope=scope)
    outputs = tf.reverse_sequence(rev_outputs, sequence_length, 0, 1)
    if not time_major:
        outputs = tf.transpose(outputs, [1, 0, 2])
    return outputs, last_state


def fused_birnn(fused_rnn, inputs, sequence_length, initial_state=(None, None), dtype=None, scope=None,
                time_major=False, backward_device=None):
    with tf.variable_scope(scope or "BiRNN"):
        sequence_length = tf.cast(sequence_length, tf.int32)
        if not time_major:
            inputs = tf.transpose(inputs, [1, 0, 2])
        outputs_fw, state_fw = fused_rnn(inputs, sequence_length=sequence_length, initial_state=initial_state[0],
                                         dtype=dtype, scope="FW")

        if backward_device is not None:
            with tf.device(backward_device):
                outputs_bw, state_bw = fused_rnn_backward(fused_rnn, inputs, sequence_length, initial_state[1], dtype,
                                                          scope="BW")
        else:
            outputs_bw, state_bw = fused_rnn_backward(fused_rnn, inputs, sequence_length, initial_state[1], dtype,
                                                      scope="BW")

        if not time_major:
            outputs_fw = tf.transpose(outputs_fw, [1, 0, 2])
            outputs_bw = tf.transpose(outputs_bw, [1, 0, 2])
    return (outputs_fw, outputs_bw), (state_fw, state_bw)


def pair_of_bidirectional_LSTMs(seq1, seq1_lengths, seq2, seq2_lengths,
                                output_size, scope=None, drop_keep_prob=1.0,
                                conditional_encoding=True):
    """Duo of bi-LSTMs over seq1 and seq2 with (optional)conditional encoding.

    Args:
        seq1 (tensor = time x batch x input): The inputs into the first biLSTM
        seq1_lengths (tensor = batch): The lengths of the sequences.
        seq2 (tensor = time x batch x input): The inputs into the second biLSTM
        seq1_lengths (tensor = batch): The lengths of the sequences.
        output_size (int): Size of the LSTMs state.
        scope (string): The TensorFlow scope for the reader.
        drop_keep_drop (float=1.0): The keep propability for dropout.

    Returns:
        Outputs (tensor): The outputs from the second bi-LSTM.
        States (tensor): The cell states from the second bi-LSTM.
    """
    with tf.variable_scope(scope or "paired_LSTM_seq1") as varscope1:
        # seq1_states: (c_fw, h_fw), (c_bw, h_bw)
        _, seq1_final_states = dynamic_bidirectional_lstm(
                        seq1, seq1_lengths, output_size, scope=varscope1,
                        drop_keep_prob=drop_keep_prob)


    with tf.variable_scope(scope or "paired_LSTM_seq2") as varscope2:
        varscope1.reuse_variables()
        # each [batch_size x max_seq_length x output_size]
        all_states_fw_bw, final_states_fw_bw = dynamic_bidirectional_lstm(
                                            seq2, seq2_lengths, output_size,
                                            seq1_final_states, scope=varscope2,
                                            drop_keep_prob=drop_keep_prob)

    return all_states_fw_bw, final_states_fw_bw


def dynamic_bidirectional_lstm(inputs, lengths, output_size,
                               initial_state=(None, None), scope=None,
                               drop_keep_prob=1.0):
    """Dynamic bi-LSTM reader, with optional initial state.

    Args:
        inputs (tensor): The inputs into the bi-LSTM
        lengths (tensor): The lengths of the sequences
        output_size (int): Size of the LSTM state of the reader.
        context (tensor=None, tensor=None): Tuple of initial
                                            (forward, backward) states
                                            for the LSTM
        scope (string): The TensorFlow scope for the reader.
        drop_keep_drop (float=1.0): The keep probability for dropout.

    Returns:
        all_states (tensor): All forward and backward states
        final_states (tensor): The final forward and backward states
    """
    with tf.variable_scope(scope or "reader"):
        cell_fw = tf.contrib.rnn.LSTMCell(
            output_size,
            state_is_tuple=True,
            initializer=tf.contrib.layers.xavier_initializer()
        )
        cell_bw = tf.contrib.rnn.LSTMCell(
            output_size,
            state_is_tuple=True,
            initializer=tf.contrib.layers.xavier_initializer()
        )

        if drop_keep_prob != 1.0:
            cell_fw = tf.contrib.rnn.DropoutWrapper(
                                    cell=cell_fw,
                                    output_keep_prob=drop_keep_prob,
                                    input_keep_prob=drop_keep_prob, seed=1233)
            cell_bw = tf.contrib.rnn.DropoutWrapper(
                                    cell=cell_bw,
                                    output_keep_prob=drop_keep_prob,
                                    input_keep_prob=drop_keep_prob, seed=1233)

        all_states_fw_bw, final_states_fw_bw = tf.nn.bidirectional_dynamic_rnn(
            cell_fw,
            cell_bw,
            inputs,
            sequence_length=lengths,
            initial_state_fw=initial_state[0],
            initial_state_bw=initial_state[1],
            dtype=tf.float32
        )

        return all_states_fw_bw, final_states_fw_bw


class SRUFusedRNN(tf.contrib.rnn.FusedRNNCell):
    """Simple Recurrent Unit, very fast.  https://openreview.net/pdf?id=rJBiunlAW"""

    def __init__(self, num_units, f_bias=1.0, r_bias=0.0, with_residual=True):
        self._num_units = num_units
        cell = _SRUUpdateCell(num_units, with_residual)
        self._rnn = tf.contrib.rnn.FusedRNNCellAdaptor(cell, use_dynamic_rnn=True)
        self._constant_bias = [0.0] * self._num_units + [f_bias] * self._num_units
        if with_residual:
            self._constant_bias += [r_bias] * self._num_units

        self._constant_bias = np.array(self._constant_bias, np.float32)
        self._with_residual = with_residual

    def __call__(self, inputs, initial_state=None, dtype=tf.float32, sequence_length=None, scope=None):
        num_gates = 3 if self._with_residual else 2
        transformed = tf.layers.dense(inputs, num_gates * self._num_units,
                                      bias_initializer=tf.constant_initializer(self._constant_bias))

        gates = tf.split(transformed, num_gates, axis=2)
        forget_gate = tf.sigmoid(gates[1])
        transformed_inputs = (1.0 - forget_gate) * gates[0]
        if self._with_residual:
            residual_gate = tf.sigmoid(gates[2])
            inputs *= (1.0 - residual_gate)
            new_inputs = tf.concat([inputs, transformed_inputs, forget_gate, residual_gate], axis=2)
        else:
            new_inputs = tf.concat([transformed_inputs, forget_gate], axis=2)

        return self._rnn(new_inputs, initial_state, dtype, sequence_length, scope)


class _SRUUpdateCell(tf.contrib.rnn.RNNCell):
    """Simple Recurrent Unit, very fast.  https://openreview.net/pdf?id=rJBiunlAW"""

    def __init__(self, num_units, with_residual, activation=None, reuse=None):
        super(_SRUUpdateCell, self).__init__(_reuse=reuse)
        self._num_units = num_units
        self._with_residual = with_residual
        self._activation = activation or tf.tanh

    @property
    def state_size(self):
        return self._num_units

    @property
    def output_size(self):
        return self._num_units

    def call(self, inputs, state):
        """Simple recurrent unit (SRU)."""
        if self._with_residual:
            base_inputs, transformed_inputs, forget_gate, residual_gate = tf.split(inputs, 4, axis=1)
            new_state = forget_gate * state + transformed_inputs
            new_h = residual_gate * self._activation(new_state) + base_inputs
        else:
            transformed_inputs, forget_gate = tf.split(inputs, 2, axis=1)
            new_state = new_h = forget_gate * state + transformed_inputs
        return new_h, new_state



import logging

import numpy as np
import tensorflow as tf

from jack.util.tf import attention, rnn
from jack.util.tf.activations import activation_from_string
from jack.util.tf.highway import highway_network

logger = logging.getLogger(__name__)


def encoder(sequence, seq_length, repr_dim=100, module='lstm', num_layers=1, reuse=None, residual=False,
            activation=None, layer_norm=False, name='encoder', dropout=None, is_eval=True, **kwargs):
    if num_layers == 1:
        if layer_norm:
            with tf.variable_scope('layernorm', reuse=False) as vs:
                vs._reuse = False  # HACK
                num_layernorms = sum(1 for v in vs.global_variables() if 'layernorm' in v.name)
                sequence = tf.contrib.layers.layer_norm(sequence, scope=str(num_layernorms))

        with tf.variable_scope(name, reuse=reuse):
            if module == 'lstm':
                out = bi_lstm(repr_dim, sequence, seq_length, **kwargs)
                if activation:
                    out = activation_from_string(activation)(out)
            elif module == 'sru':
                with_residual = sequence.get_shape()[2].value == repr_dim
                out = bi_sru(repr_dim, sequence, seq_length, with_residual, **kwargs)
                if activation:
                    out = activation_from_string(activation)(out)
            elif module == 'rnn':
                out = bi_rnn(repr_dim, tf.nn.rnn_cell.BasicRNNCell(repr_dim, activation_from_string(activation)),
                             sequence, seq_length, **kwargs)
            elif module == 'gru':
                out = bi_rnn(repr_dim, tf.contrib.rnn.GRUBlockCell(repr_dim), sequence,
                             seq_length, **kwargs)
                if activation:
                    out = activation_from_string(activation)(out)
            elif module == 'gldr':
                out = gated_linear_dilated_residual_network(
                    repr_dim, sequence, **kwargs)
            elif module == 'conv':
                out = convnet(repr_dim, sequence, 1, activation=activation_from_string(activation), **kwargs)
            elif module == 'conv_glu':
                out = gated_linear_convnet(repr_dim, sequence, 1, **kwargs)
            elif module == 'conv_separable':
                out = depthwise_separable_convolution(
                    repr_dim, sequence, activation=activation_from_string(activation), **kwargs)
            elif module == 'dense':
                out = tf.layers.dense(sequence, repr_dim)
                if activation:
                    out = activation_from_string(activation)(out)
            elif module == 'highway':
                out = highway_network(sequence, num_layers, activation_from_string(activation))
            elif module == 'self_attn':
                outs = []
                num_attn = kwargs.get('num_attn_heads', 1)
                for i in range(num_attn):
                    with tf.variable_scope(str(i)):
                        attn = self_attention(sequence, seq_length, repr_dim=repr_dim, **kwargs)
                        outs.append(attn)
                out = tf.concat(outs, 2) if num_attn > 1 else outs[0]
            elif module == 'positional_encoding':
                out = positional_encoding(sequence, seq_length)
            else:
                raise ValueError("Unknown encoder type: %s" % module)

            if residual:
                if out.get_shape()[-1].value != sequence.get_shape()[-1].value:
                    logging.error(
                        'Residual connection only possible if input to sequence encoder %s of type %s has same '
                        'dimension (%d) as output (%d).' % (name, module, sequence.get_shape()[-1].value,
                                                            out.get_shape()[-1].value))
                    raise RuntimeError()
                out += sequence

            if dropout is not None:
                out = tf.cond(
                    tf.logical_and(tf.greater(dropout, 0.0), tf.logical_not(is_eval)),
                    lambda: tf.nn.dropout(out, 1.0 - dropout, noise_shape=[tf.shape(out)[0], 1, tf.shape(out)[-1]]),
                    lambda: out)
    else:
        out = encoder(sequence, seq_length, repr_dim, module, num_layers - 1, reuse, residual,
                      activation, layer_norm, name, dropout=dropout, is_eval=is_eval, **kwargs)

        out = encoder(out, seq_length, repr_dim, module, 1, reuse, residual, activation, layer_norm,
                      name + str(num_layers - 1), dropout=dropout, is_eval=is_eval, **kwargs)

    return out


# RNN Encoders
def _bi_rnn(size, fused_rnn, sequence, seq_length, with_projection=False):
    output = rnn.fused_birnn(fused_rnn, sequence, seq_length, dtype=tf.float32, scope='rnn')[0]
    output = tf.concat(output, 2)
    if with_projection:
        projection_initializer = tf.constant_initializer(np.concatenate([np.eye(size), np.eye(size)]))
        output = tf.layers.dense(output, size, kernel_initializer=projection_initializer, name='projection')
    return output


def bi_lstm(size, sequence, seq_length, with_projection=False, **kwargs):
    fused_rnn = tf.contrib.rnn.LSTMBlockFusedCell(size)
    return _bi_rnn(size, fused_rnn, sequence, seq_length, with_projection)


def bi_rnn(size, rnn_cell, sequence, seq_length, with_projection=False, **kwargs):
    fused_rnn = tf.contrib.rnn.FusedRNNCellAdaptor(rnn_cell, use_dynamic_rnn=True)
    return _bi_rnn(size, fused_rnn, sequence, seq_length, with_projection)


def bi_sru(size, sequence, seq_length, with_residual=True, name='bi_sru', reuse=None, with_projection=False, **kwargs):
    """Simple Recurrent Unit, very fast.  https://openreview.net/pdf?id=rJBiunlAW."""
    fused_rnn = rnn.SRUFusedRNN(size, with_residual)
    return _bi_rnn(size, fused_rnn, sequence, seq_length, with_projection)


# Convolution Encoders


def convnet(repr_dim, inputs, num_layers, conv_width=3, activation=tf.nn.relu, **kwargs):
    # dim reduction
    output = inputs
    for i in range(num_layers):
        output = _convolutional_block(output, repr_dim, conv_width=conv_width, name="conv_%d" % i)
    return output


def _convolutional_block(inputs, out_channels, conv_width=3, name='conv', activation=tf.nn.relu, **kwargs):
    channels = inputs.get_shape()[2].value
    # [filter_height, filter_conv_width, in_channels, out_channels]
    f = tf.get_variable(name + '_filter', [conv_width, channels, out_channels])
    output = tf.nn.conv1d(inputs, f, 1, padding='SAME', name=name)
    return activation(output)


def depthwise_separable_convolution(repr_dim, inputs, conv_width, activation=tf.nn.relu, bias=True, **kwargs):
    inputs = tf.expand_dims(inputs, 1)
    shapes = inputs.shape.as_list()

    depthwise_filter = tf.get_variable("depthwise_filter",
                                       (1, conv_width, shapes[-1], 1),
                                       dtype=tf.float32)
    pointwise_filter = tf.get_variable("pointwise_filter", (1, 1, shapes[-1], repr_dim),
                                       dtype=tf.float32)
    outputs = tf.nn.separable_conv2d(inputs,
                                     depthwise_filter,
                                     pointwise_filter,
                                     strides=(1, 1, 1, 1),
                                     padding="SAME")
    outputs = tf.squeeze(outputs, 1)
    if bias:
        b = tf.get_variable("bias", outputs.shape[-1], initializer=tf.zeros_initializer())
        outputs += b
    outputs = activation(outputs)
    return outputs


# following implementation of fast encoding in https://openreview.net/pdf?id=HJRV1ZZAW
def _residual_dilated_convolution_block(inputs, dilation=1, conv_width=3, name="dilated_conv"):
    # [filter_height, filter_conv_width, in_channels, out_channels]
    output = inputs
    channels = inputs.get_shape()[2].value
    for i in range(2):
        # [filter_height, filter_conv_width, in_channels, out_channels]
        output = _convolutional_glu_block(output, channels, dilation, conv_width, name=name + '_' + str(i))
    return output + inputs


def _convolutional_glu_block(inputs, out_channels, dilation=1, conv_width=3, name='conv_glu', **kwargs):
    channels = inputs.get_shape()[2].value
    # [filter_height, filter_conv_width, in_channels, out_channels]
    f = tf.get_variable(name + '_filter', [1, conv_width, channels, out_channels * 2])
    output = tf.nn.atrous_conv2d(tf.expand_dims(inputs, 1), f, dilation, 'SAME', name=name)
    output = tf.squeeze(output, 1)
    output, gate = tf.split(output, 2, 2)
    output *= tf.sigmoid(gate)
    return output


def gated_linear_dilated_residual_network(out_size, inputs, dilations, conv_width=3, name='gldr_network', reuse=None,
                                          **kwargs):
    """Follows https://openreview.net/pdf?id=HJRV1ZZAW.

    Args:
        out_size: size of output
        inputs: input sequence tensor [batch_size, length, size]
        dilations: list, representing (half of the) network depth; each residual dilation block comprises 2 convolutions

    Returns:
        [batch_size, length, out_size] tensor
    """
    # dim reduction
    output = _convolutional_glu_block(inputs, out_size, name='conv_dim_reduction')
    for i, d in enumerate(dilations):
        output = _residual_dilated_convolution_block(output, d, conv_width, name='dilated_conv_%d' % i)
    return output


def gated_linear_convnet(out_size, inputs, num_layers, conv_width=3, **kwargs):
    """Follows https://openreview.net/pdf?id=HJRV1ZZAW.

    Args:
        out_size: size of output
        inputs: input sequence tensor [batch_size, length, size]
        num_layers: number of conv layers with conv_width

    Returns:
        [batch_size, length, out_size] tensor
    """
    # dim reduction
    output = inputs
    for i in range(num_layers):
        output = _convolutional_glu_block(output, out_size, conv_width=conv_width, name="conv_%d" % i)
    return output


def positional_encoding(inputs, lengths, **kwargs):
    repr_dim = inputs.get_shape()[-1].value
    pos = tf.reshape(tf.range(0.0, tf.to_float(tf.reduce_max(lengths)), dtype=tf.float32), [-1, 1])
    i = np.arange(0, repr_dim, 2, np.float32)
    denom = np.reshape(np.power(10000.0, i / repr_dim), [1, -1])
    enc = tf.expand_dims(tf.concat([tf.sin(pos / denom), tf.cos(pos / denom)], 1), 0)
    return inputs + tf.tile(enc, [tf.shape(inputs)[0], 1, 1])


# Self attention layers
def self_attention(inputs, lengths, attn_type='bilinear', scaled=True, activation=None, with_sentinel=False, **kwargs):
    if attn_type == 'bilinear':
        attn_states = attention.bilinear_attention(
            inputs, inputs, lengths, scaled, with_sentinel, **kwargs)[2]
    elif attn_type == 'dot':
        attn_states = attention.dot_attention(
            inputs, inputs, lengths, scaled, with_sentinel, **kwargs)[2]
    elif attn_type == 'diagonal_bilinear':
        attn_states = \
            attention.diagonal_bilinear_attention(
                inputs, inputs, lengths, scaled, with_sentinel, **kwargs)[2]
    elif attn_type == 'mlp':
        attn_states = \
            attention.mlp_attention(
                kwargs['repr_dim'], activation, inputs, inputs, lengths, with_sentinel, **kwargs)[2]
    else:
        raise ValueError("Unknown attention type: %s" % attn_type)

    return attn_states



"""Tensor ports are used to define 'module signatures'

Modules are loosly coupled with each other via the use of tensor ports. They simply define what kind of tensors are
produced at the input and/or output of each module, thus defining a kind of signature. This allows for maximum
flexibility when (re-)using modules in different combinations.
"""
import logging
from typing import Mapping, Sequence, Any

import numpy as np
import tensorflow as tf

try:
    import torch
except ImportError as e:
    pass

logger = logging.getLogger(__name__)

_allowed_torch = {np.dtype(np.int8), np.dtype(np.int16), np.dtype(np.int32), np.dtype(np.int64),
                  np.dtype(np.float16), np.dtype(np.float32), np.dtype(np.float64)}


class TensorPort:
    """A TensorPort defines an input or output tensor for a modules.

    A port defines at least a shape, name, and its data type.
    """

    def __init__(self, dtype, shape, name, doc_string=None, shape_string=None):
        """Create a new TensorPort.

        Args:
            dtype: the (numpy) data type of the port.
            shape: the shape of the tensor.
            name: the name of this port (should be a valid TF name)
            doc_string: a documentation string associated with this port
            shape_string: a string of the form [size_1,size_2,size_3] where size_i is a text describing the
                size of the tensor's dimension i (such as "number of batches").
        """
        self.dtype = np.dtype(dtype)
        self.shape = shape
        self.name = name
        self.__doc__ = doc_string
        self.shape_string = shape_string

    def create_tf_placeholder(self):
        """Convenience method that produces a placeholder of the type and shape defined by the port.

        Returns: a placeholder of same type, shape and name.
        """
        return tf.placeholder(tf.as_dtype(self.dtype), self.shape, self.name)

    def create_torch_variable(self, value, gpu=False):
        """Convenience method that produces a tensor given the value of the defined type.

        Returns: a torch tensor of same type.
        """
        if isinstance(value, torch.autograd.Variable):
            if gpu:
                value = value.cuda()
            return value
        if not torch.is_tensor(value):
            if not isinstance(value, np.ndarray):
                value = np.array(value, dtype=self.dtype)
            else:
                value = value.astype(self.dtype)
            if value.size == 0:
                return value

            if self.dtype in _allowed_torch:
                value = torch.autograd.Variable(torch.from_numpy(value))
        else:
            value = torch.autograd.Variable(value)
        if gpu and isinstance(value, torch.autograd.Variable):
            value = value.cuda()
        return value

    @staticmethod
    def torch_to_numpy(value):
        """Convenience method that produces a tensor given the value of the defined type.

        Returns: a torch tensor of same type.
        """
        if isinstance(value, torch.autograd.Variable):
            value = value.data
        if torch.is_tensor(value):
            return value.cpu().numpy()
        elif isinstance(value, np.ndarray):
            return value
        else:
            return np.ndarray(value)

    def get_description(self):
        """Returns a multi-line description string of the TensorPort."""

        return "Tensorport '%s'" % self.name + "\n" + \
               "  dtype: " + str(self.dtype) + "\n" + \
               "  shape: " + str(self.shape) + "\n" + \
               "  doc_string: " + str(self.__doc__) + "\n" + \
               "  shape_string: " + str(self.shape_string)

    def __gt__(self, port):
        return self.name > port.name

    def __repr__(self):
        return "<TensorPort (%s)>" % self.name

    @staticmethod
    def to_mapping(ports: Sequence['TensorPort'], tensors: Sequence[tf.Tensor]):
        """
        Create a dictionary of ports to tensors based on ordered port and tensor sequences.
        Args:
            ports: list of ports
            tensors: list of tensors (same length as ports)

        Returns: mapping from the i-th port to the i-th tensor in the lists.

        """
        return dict(zip(ports, tensors))


class TensorPortWithDefault(TensorPort):
    """
    TensorPort that also defines a default value.
    """

    def __init__(self, default_value, shape, name, doc_string=None, shape_string=None):
        """Default value must be a numpy array."""
        self.default_value = default_value
        super().__init__(default_value.dtype, shape, name, doc_string=doc_string, shape_string=shape_string)

    def create_tf_placeholder(self):
        """Creates a TF placeholder_with_default.

        Convenience method that produces a constant of the type, value and shape defined by the port.
        Returns: a constant tensor of same type, shape and name. It can nevertheless be fed with external values
        as if it was a placeholder.
        """
        ph = tf.placeholder_with_default(self.default_value, self.shape, self.name)
        if ph.dtype != tf.as_dtype(self.dtype):
            logger.warning(
                "Placeholder {} with default of type {} created for TensorPort with type {}!".format(self.name,
                                                                                                     ph.dtype,
                                                                                                     self.dtype))
        return ph

    def create_torch_variable(self, value, gpu=False):
        if value is None:
            value = self.default_value
        return super(TensorPortWithDefault, self).create_torch_variable(value, gpu)


class Ports:
    """Defines sopme common ports for reusability and as examples. Readers can of course define their own.

    This class groups input ports. Different modules can refer to these ports
    to define their input or output, respectively.
    """

    loss = TensorPort(np.float32, [None], "loss",
                      "Represents loss on each instance in the batch",
                      "[batch_size]")
    keep_prob = TensorPortWithDefault(np.array(1.0, np.float32), [], "keep_prob",
                                      "scalar representing keep probability when using dropout",
                                      "[]")
    is_eval = TensorPortWithDefault(np.array(True), [], "is_eval",
                                    "boolean that determines whether input is eval or training.",
                                    "[]")

    class Input:
        question = TensorPort(np.int32, [None, None], "question",
                              "Represents questions using symbol vectors",
                              "[batch_size, max_num_question_tokens]")

        support = TensorPort(np.int32, [None, None], "support",
                             "Represents instances with single support documents",
                             "[batch_size, max_num_tokens]")

        multiple_support = TensorPort(np.int32, [None, None, None], "multiple_support",
                                      ("Represents instances with multiple support documents",
                                       " or single instances with extra dimension set to 1"),
                                      "[batch_size, max_num_support, max_num_tokens]")

        atomic_candidates = TensorPort(np.int32, [None, None], "atomic_candidates",
                                       ("Represents candidate choices using single symbols. ",
                                        "This could be a list of entities from global entities ",
                                        "for example atomic_candidates = [e1, e7, e83] from ",
                                        "global_entities = [e1, e2, e3, ..., eN-1, eN"),
                                       "[batch_size, num_candidates]")

        sample_id = TensorPort(np.int32, [None], "sample_id",
                               "Maps this sample to the index in the input data",
                               "[batch_size]")

        muliple_support_length = TensorPort(np.int32, [None, None], "muliple_support_length",
                                            "Represents length of supports in each support in batch",
                                            "[batch_size, num_supports]")

        support_length = TensorPort(np.int32, [None], "support_length",
                                    "Represents length of supports in batch",
                                    "[batch_size]")

        question_length = TensorPort(np.int32, [None], "question_length",
                                     "Represents length of questions in batch",
                                     "[batch_size]")

        emb_support = TensorPort(np.float32, [None, None, None], "emb_support",
                                      "Represents the embedded support",
                                      "[S, max_num_tokens, N]")

        emb_question = TensorPort(np.float32, [None, None, None], "emb_question",
                                       "Represents the embedded question",
                                       "[Q, max_num_question_tokens, N]")

        # character based information
        word_chars = TensorPort(np.int32, [None, None], "word_chars",
                                "Represents questions using symbol vectors",
                                "[U, max_num_chars]")
        word_char_length = TensorPort(np.int32, [None], "word_char_length",
                                      "Represents questions using symbol vectors",
                                      "[U]")
        question_batch_words = TensorPort(np.int32, [None, None], "question_batch_words",
                                          "Represents question using in-batch vocabulary.",
                                          "[batch_size, max_num_question_tokens]")
        support_batch_words = TensorPort(np.int32, [None, None], "support_batch_words",
                                         "Represents support using in-batch vocabulary",
                                         "[batch_size, max_num_support_tokens]")

        # Number of questions in batch is Q, number of supports is S, number of answers is A, number of candidates is C.
        # Typical input ports such as support, candidates, answers are defined together with individual mapping ports.
        # This allows for more flexibility when numbers can vary between questions.

        support2question = TensorPort(np.int32, [None], "support2question",
                                      "Represents mapping to question idx per support",
                                      "[S]")
        candidate2question = TensorPort(np.int32, [None], "candidate2question",
                                        "Represents mapping to question idx per candidate",
                                        "[C]")
        answer2support = TensorPortWithDefault(np.array([0], np.int32), [None], "answer2support",
                                               "Represents mapping to support idx per answer", "[A]")
        atomic_candidates1D = TensorPort(np.int32, [None], "candidates1D",
                                         "Represents candidate choices using single symbols",
                                         "[C]")

        seq_candidates = TensorPort(np.int32, [None, None], "seq_candidates",
                                    "Represents candidate choices using single symbols",
                                    "[C, max_num_tokens]")

        # MISC intermediate ports that might come in handy
        # -embeddings
        embedded_seq_candidates = TensorPort(np.float32, [None, None, None], "embedded_seq_candidates_flat",
                                             "Represents the embedded sequential candidates",
                                             "[C, max_num_tokens, N]")

        embedded_candidates = TensorPort(np.float32, [None, None], "embedded_candidates_flat",
                                         "Represents the embedded candidates",
                                         "[C, N]")

    class Prediction:
        logits = TensorPort(np.float32, [None, None], "logits",
                            "Represents output scores for each candidate",
                            "[C, num_candidates]")

        candidate_index = TensorPort(np.float32, [None], "candidate_idx",
                                     "Represents answer as a single index",
                                     "[C]")

        candidate_scores = TensorPort(np.float32, [None], "candidate_scores_flat",
                                      "Represents output scores for each candidate",
                                      "[C]")

        candidate_idx = TensorPort(np.float32, [None], "candidate_predictions_flat",
                                   "Represents groundtruth candidate labels, usually 1 or 0",
                                   "[C]")

        # extractive QA
        start_scores = TensorPort(np.float32, [None, None], "start_scores",
                                  "Represents start scores for each support sequence",
                                  "[S, max_num_tokens]")

        end_scores = TensorPort(np.float32, [None, None], "end_scores",
                                "Represents end scores for each support sequence",
                                "[S, max_num_tokens]")

        answer_span = TensorPort(np.int32, [None, 3], "answer_span",
                                 "Represents answer as a (doc_idx, start, end) span", "[A, 3]")

        # generative QA
        generative_symbol_scores = TensorPort(np.int32, [None, None, None], "symbol_scores",
                                              "Represents symbol scores for each possible "
                                              "sequential answer given during training",
                                              "[A, max_num_tokens, vocab_len]")

        generative_symbols = TensorPort(np.int32, [None, None], "symbol_prediction",
                                        "Represents symbol sequence for each possible "
                                        "answer target_indexpredicted by the model",
                                        "[A, max_num_tokens]")

    class Target:
        candidate_1hot = TensorPort(np.float32, [None, None], "candidate_targets",
                                    "Represents target (0/1) values for each candidate",
                                    "[batch_size, num_candidates]")

        target_index = TensorPort(np.int32, [None], "target_index",
                                  ("Represents symbol id of target candidate. ",
                                   "This can either be an index into a full list of candidates,",
                                   " which is fixed, or an index into a partial list of ",
                                   "candidates, for example a list of potential entities ",
                                   "from a list of many candidates"),
                                  "[batch_size]")

        answer_span = TensorPort(np.int32, [None, 2], "answer_span_target",
                                 "Represents answer as a (start, end) span", "[A, 2]")

        seq_answer = TensorPort(np.int32, [None, None], "answer_seq_target",
                                "Represents answer as a sequence of symbols",
                                "[A, max_num_tokens]")

        symbols = TensorPort(np.int32, [None, None], "symbol_targets",
                             "Represents symbols for each possible target answer sequence",
                             "[A, max_num_tokens]")


class TensorPortTensors:
    """
    This class wraps around mappings from tensor ports to tensors and makes the tensors available by
    by `x.foo` instead of `x['foo']` calls.
    """

    def __init__(self, mapping: Mapping[TensorPort, Any]):
        """
        Create a wrapping based on the passed in mapping/dictionary.
        Args:
            mapping: Mapping from ports to tensors.
        """
        self.name_to_tensor = ({key.name: value for key, value in mapping.items()})

    def __getattr__(self, item):
        """
        Returns the tensor belonging to the port with the given name.
        Args:
            item: the tensor port name.

        Returns: the tensor associated with the tensor port of the given name.
        """
        return self.name_to_tensor[item]



import torch
from torch import nn


class BiLSTM(nn.Module):
    def __init__(self, input_size, size, start_state_given=False):
        super(BiLSTM, self).__init__()
        self._size = size
        self._bilstm = nn.LSTM(input_size, size, 1, bidirectional=True, batch_first=True)
        self._bilstm.bias_ih_l0.data[size:2 * size].fill_(1.0)
        self._bilstm.bias_ih_l0_reverse.data[size:2 * size].fill_(1.0)
        self._start_state_given = start_state_given
        if not start_state_given:
            self._lstm_start_hidden = nn.Parameter(torch.zeros(2, size))
            self._lstm_start_state = nn.Parameter(torch.zeros(2, size))

    def forward(self, inputs, lengths=None, start_state=None):
        if not self._start_state_given:
            batch_size = inputs.size(0)
            start_hidden = self._lstm_start_hidden.unsqueeze(1).expand(2, batch_size, self._size).contiguous()
            start_state = self._lstm_start_state.unsqueeze(1).expand(2, batch_size, self._size).contiguous()
            start_state = (start_hidden, start_state)

        if lengths is not None:
            new_lengths, indices = torch.sort(lengths, dim=0, descending=True)
            inputs = torch.index_select(inputs, 0, indices)
            if self._start_state_given:
                start_state = (torch.index_select(start_state[0], 1, indices),
                               torch.index_select(start_state[1], 1, indices))
            new_lengths = [l.data[0] for l in new_lengths]
            inputs = nn.utils.rnn.pack_padded_sequence(inputs, new_lengths, batch_first=True)

        output, (h_n, c_n) = self._bilstm(inputs, start_state)

        if lengths is not None:
            output = nn.utils.rnn.pad_packed_sequence(output, batch_first=True)[0]
            _, back_indices = torch.sort(indices, dim=0)
            output = torch.index_select(output, 0, back_indices)
            h_n = torch.index_select(h_n, 1, back_indices)
            c_n = torch.index_select(c_n, 1, back_indices)

        return output, (h_n, c_n)



import tensorflow as tf

from jack.readers.classification.shared import AbstractSingleSupportClassificationModel
from jack.util.tf.rnn import fused_birnn


class ConditionalBiLSTMClassificationModel(AbstractSingleSupportClassificationModel):
    def forward_pass(self, shared_resources, embedded_question, embedded_support, num_classes, tensors):
        # question - hypothesis; support - premise
        repr_dim = shared_resources.config['repr_dim']
        dropout = shared_resources.config.get("dropout", 0.0)

        with tf.variable_scope('embedding_projection') as vs:
            embedded_question = tf.layers.dense(embedded_question, repr_dim, tf.tanh, name='projection')
            vs.reuse_variables()
            embedded_support = tf.layers.dense(embedded_support, repr_dim, tf.tanh, name='projection')
            # keep dropout mask constant over time
            dropout_shape = [tf.shape(embedded_question)[0], 1, tf.shape(embedded_question)[2]]
            embedded_question = tf.nn.dropout(embedded_question, 1.0 - dropout, dropout_shape)
            embedded_support = tf.nn.dropout(embedded_support, 1.0 - dropout, dropout_shape)

        fused_rnn = tf.contrib.rnn.LSTMBlockFusedCell(repr_dim)
        # [batch, 2*output_dim] -> [batch, num_classes]
        _, q_states = fused_birnn(fused_rnn, embedded_question, sequence_length=tensors.question_length,
                                  dtype=tf.float32, time_major=False, scope="question_rnn")

        outputs, _ = fused_birnn(fused_rnn, embedded_support, sequence_length=tensors.support_length,
                                 dtype=tf.float32, initial_state=q_states, time_major=False, scope="support_rnn")

        # [batch, T, 2 * dim] -> [batch, dim]
        outputs = tf.concat([outputs[0], outputs[1]], axis=2)
        hidden = tf.layers.dense(outputs, repr_dim, tf.nn.relu, name="hidden") * tf.expand_dims(
            tf.sequence_mask(tensors.support_length, maxlen=tf.shape(outputs)[1], dtype=tf.float32), 2)
        hidden = tf.reduce_max(hidden, axis=1)
        # [batch, dim] -> [batch, num_classes]
        outputs = tf.layers.dense(hidden, num_classes, name="classification")
        return outputs



# -*- coding: utf-8 -*-

import sys

import tensorflow as tf


def hinge_loss(positive_scores, negative_scores, margin=1.0):
    """
    Pairwise hinge loss [1]:
        loss(p, n) = \sum_i [\gamma - p_i + n_i]_+

    [1] http://yann.lecun.com/exdb/publis/pdf/lecun-06.pdf

    Args:
        positive_scores: (N,) Tensor containing scores of positive examples.
        negative_scores: (N,) Tensor containing scores of negative examples.
        margin: Margin.
    Returns:
        Loss value.
    """
    hinge_losses = tf.nn.relu(margin - positive_scores + negative_scores)
    loss = tf.reduce_sum(hinge_losses)
    return loss


def logistic_loss(positive_scores, negative_scores):
    """
    Pairwise logistic loss [1]:
        loss(p, n) = \sum_i log(1 + e^(1 - p_i + n_i))

    [1] http://yann.lecun.com/exdb/publis/pdf/lecun-06.pdf

    Args:
        positive_scores: (N,) Tensor containing scores of positive examples.
        negative_scores: (N,) Tensor containing scores of negative examples.
    Returns:
        Loss value.
    """
    logistic_losses = tf.log(1 + tf.exp(1 - positive_scores + negative_scores))
    loss = tf.reduce_sum(logistic_losses)
    return loss


def mce_loss(positive_scores, negative_scores):
    """
    Minimum Classification Error (MCE) loss [1]:
        loss(p, n) = \sum_i \sigma(- p_i + n_i)

    [1] http://yann.lecun.com/exdb/publis/pdf/lecun-06.pdf

    Args:
        positive_scores: (N,) Tensor containing scores of positive examples.
        negative_scores: (N,) Tensor containing scores of negative examples.
    Returns:
        Loss value.
    """
    mce_losses = tf.sigmoid(- positive_scores + negative_scores)
    loss = tf.reduce_sum(mce_losses)
    return loss


def square_square_loss(positive_scores, negative_scores, margin=1.0):
    """
    Square-Square loss [1]:
        loss(p, n) = \sum_i - p_i^2 + [\gamma + n_i]^2_+

    [1] http://yann.lecun.com/exdb/publis/pdf/lecun-06.pdf

    Args:
        positive_scores: (N,) Tensor containing scores of positive examples.
        negative_scores: (N,) Tensor containing scores of negative examples.
    :   margin: Margin.
    Returns:
        Loss value.
    """
    square_square_losses = - positive_scores + tf.nn.relu(margin + negative_scores) ** 2
    loss = tf.reduce_sum(square_square_losses)
    return loss


def square_exponential_loss(positive_scores, negative_scores, gamma=1.0):
    """
    Square-Exponential loss [1]:
        loss(p, n) = \sum_i - p_i^2 + \gamma e^(n_i)

    [1] http://yann.lecun.com/exdb/publis/pdf/lecun-06.pdf

    Args:
        positive_scores: (N,) Tensor containing scores of positive examples.
        negative_scores: (N,) Tensor containing scores of negative examples.
        gamma: Gamma hyper-parameter.
    Returns:
        Loss value.
    """
    square_exponential_losses = - positive_scores + gamma * tf.exp(negative_scores)
    loss = tf.reduce_sum(square_exponential_losses)
    return loss


def get_function(function_name):
    this_module = sys.modules[__name__]
    if not hasattr(this_module, function_name):
        raise ValueError('Unknown objective function: {}'.format(function_name))
    return getattr(this_module, function_name)




#!/usr/bin/env python3

import json
import jsonschema
from sys import argv

def main(arg1, arg2):
    with open(arg1) as f:
        data = json.load(f)

    with open(arg2) as f:
        schema = json.load(f)

    try:
        jsonschema.validate(data, schema)
        return 'JSON successfully validated.'
    except jsonschema.ValidationError as e:
        return e.message
    except jsonschema.SchemaError as e:
        return e


if __name__ == '__main__':
    response = main(argv[1], argv[2])
    print(response)



"""
This script converts data from the SemEval-2007 Task 10 on English Lexical
Substitution to the jtr format.
"""

import json
import xmltodict
import re
import os


def load_substitituons(path):
    subs = {}
    with open(path, "r") as f:
        for line in f.readlines()[1:]:
            splits = line.split(" :: ")
            id = splits[0].split(" ")[1]
            sub = [x[:-2] for x in splits[1].split(";")][:-1]
            print(id, sub)
            subs[id] = sub
    return subs


if __name__ == "__main__":
    CLOZE_STYLE = False

    for corpus_name in ["trial"]:
        file_path = "./jack/data/LS/%s/lexsub_%s_cleaned.xml" % (
        corpus_name, corpus_name)
        subs_path = "./jack/data/LS/%s/gold.%s" % (corpus_name, corpus_name)
        subs = load_substitituons(subs_path)

        with open(file_path) as fd:
            file_text = fd.read().replace("&", "&#038;")
            corpus = xmltodict.parse(file_text)["corpus"]

            jtr = {"meta": "SemEval-2007 Task 10: Lexical Substitution"}

            instances = []

            for lexelt in corpus["lexelt"]:
                for instance in lexelt["instance"]:
                    # fixme: not sure what is happening here
                    if str(instance) != "@id" and str(instance) != "context":
                        context = instance["context"]
                        id = instance["@id"]
                        word = re.search('_START_\w+_END_', context).group(0)[
                               7:-5]
                        context_masked = re.sub('_START_\w+_END_', 'XXXXX',
                                                context)
                        context_recovered = re.sub('_START_\w+_END_', word,
                                                   context)
                        context_tokenized = context.split(" ")
                        word_position = [i for i, word in enumerate(context_tokenized) if word.startswith('_START_')][0]

                        # print("%s\t%s\t%s" % (id, word, context_masked))

                        if CLOZE_STYLE:
                            queb = {'id': id, 'support': [], 'questions': [
                                {'question': context_masked,
                                 'answers': [
                                     {'text': word}
                                 ]}
                            ]}
                        else:
                            queb = {'id': id,
                                    'support': [{'text': context_recovered}],
                                    'questions': [
                                        {'question': str(word_position),
                                         'answers': [
                                             {'text': sub} for sub in subs[id]
                                         ]}
                                    ]}

                        instances.append(queb)

            jtr["instances"] = instances

            with open("./jack/data/LS/%s/lexsub_%s_cleaned.jsonl" % \
                              (corpus_name, corpus_name), 'w') as outfile:
                json.dump(jtr, outfile, indent=2)

            # create snippet
            jtr['instances'] = jtr['instances'][:10]


            def save_debug(directory_path, file_name):
                if not os.path.exists(directory_path):
                    os.makedirs(directory_path)

                with open(directory_path + "/" + file_name, 'w') as outfile:
                    json.dump(jtr, outfile, indent=2)


            save_debug("./data/LS/debug", "lexsub_debug_cleaned.jsonl")
            save_debug("./data/LS", "snippet.jack.json")



# -*- coding: utf-8 -*-

import tensorflow as tf


def highway_layer(inputs, activation, name=None):
    with tf.variable_scope(name or "highway_layer"):
        d = inputs.get_shape()[-1].value
        trans_gate = tf.contrib.layers.fully_connected(inputs, 2 * d, activation_fn=None, weights_initializer=None,
                                                       scope='trans_gate')
        trans, gate = tf.split(trans_gate, 2, len(inputs.get_shape()) - 1)
        trans, gate = activation(trans), tf.sigmoid(gate)
        out = gate * trans + (1 - gate) * inputs
        return out


def highway_network(inputs, num_layers, activation=tf.tanh, name=None, reuse=False):
    with tf.variable_scope(name or "highway_network", reuse=reuse):
        prev = inputs
        cur = None
        for layer_idx in range(num_layers):
            cur = highway_layer(prev, activation, name="layer_{}".format(layer_idx))
            prev = cur
    return cur



# -*- coding: utf-8 -*-

from typing import NamedTuple

from jack.core import *
from jack.core.data_structures import *
from jack.core.tensorflow import TFModelModule
from jack.readers.classification import util
from jack.util import preprocessing
from jack.util.map import numpify

logger = logging.getLogger(__name__)


class AbstractSingleSupportClassificationModel(TFModelModule):
    def __init__(self, shared_resources):
        self.shared_resources = shared_resources
        self.vocab = self.shared_resources.vocab
        self.config = self.shared_resources.config

        super(AbstractSingleSupportClassificationModel, self).__init__(shared_resources)

    @property
    def input_ports(self) -> List[TensorPort]:
        if self.shared_resources.embeddings is not None:
            return [Ports.Input.emb_support, Ports.Input.emb_question,
                    Ports.Input.support, Ports.Input.question,
                    Ports.Input.support_length, Ports.Input.question_length,
                    # character information
                    Ports.Input.word_chars, Ports.Input.word_char_length,
                    Ports.Input.question_batch_words, Ports.Input.support_batch_words,
                    Ports.is_eval]
        else:
            return [Ports.Input.support, Ports.Input.question,
                    Ports.Input.support_length, Ports.Input.question_length,
                    # character information
                    Ports.Input.word_chars, Ports.Input.word_char_length,
                    Ports.Input.question_batch_words, Ports.Input.support_batch_words, Ports.is_eval]

    @property
    def output_ports(self) -> List[TensorPort]:
        return [Ports.Prediction.logits, Ports.Prediction.candidate_index]

    @property
    def training_input_ports(self) -> List[TensorPort]:
        return [Ports.Prediction.logits, Ports.Target.target_index]

    @property
    def training_output_ports(self) -> List[TensorPort]:
        return [Ports.loss]

    def create_output(self, shared_resources: SharedResources, input_tensors) -> Mapping[TensorPort, tf.Tensor]:
        tensors = TensorPortTensors(input_tensors)
        embedded_question = []
        embedded_support = []

        input_size = shared_resources.config.get('repr_dim_task_embedding', 0)

        if input_size > 0:
            vocab_size = len(shared_resources.vocab)
            e = tf.get_variable('embeddings', [vocab_size, input_size],
                                tf.float32, initializer=tf.random_normal_initializer(0.0, 0.1), trainable=True)
            embedded_question.append(tf.nn.embedding_lookup(e, input_tensors[Ports.Input.question]))
            embedded_support.append(tf.nn.embedding_lookup(e, input_tensors[Ports.Input.support]))

        if shared_resources.embeddings is not None:
            embedded_support.append(input_tensors[Ports.Input.emb_support])
            embedded_question.append(input_tensors[Ports.Input.emb_question])
            input_size += shared_resources.embeddings.shape[-1]

        if len(embedded_question) > 1:
            embedded_question = tf.concat(embedded_question, 2)
            embedded_support = tf.concat(embedded_support, 2)
        else:
            embedded_question = embedded_question[0]
            embedded_support = embedded_support[0]

        embedded_question.set_shape([None, None, input_size])
        embedded_support.set_shape([None, None, input_size])

        logits = self.forward_pass(shared_resources, embedded_question, embedded_support,
                                   len(self.shared_resources.answer_vocab), tensors)

        predictions = tf.argmax(logits, 1, name='prediction')

        return {
            Ports.Prediction.logits: logits,
            Ports.Prediction.candidate_index: predictions
        }

    @abstractmethod
    def forward_pass(self, shared_resources, embedded_question, embedded_support, num_classes, tensors):
        """Takes embedded support and question and produces logits"""
        raise NotImplementedError

    def create_training_output(self, shared_resources: SharedResources, input_tensors):
        tensors = TensorPortTensors(input_tensors)
        return {
            Ports.loss: tf.losses.sparse_softmax_cross_entropy(logits=tensors.logits, labels=tensors.target_index)
        }


MCAnnotation = NamedTuple('MCAnnotation', [
    ('question_tokens', List[str]),
    ('question_ids', List[int]),
    ('question_length', int),
    ('support_tokens', List[str]),
    ('support_ids', List[int]),
    ('support_length', int),
    ('answer', Optional[int]),
    ('id', Optional[int]),
])


class ClassificationSingleSupportInputModule(OnlineInputModule[MCAnnotation]):
    def setup(self):
        self.vocab = self.shared_resources.vocab
        self.config = self.shared_resources.config
        self.embeddings = self.shared_resources.embeddings
        if self.embeddings is not None:
            self.__default_vec = np.zeros([self.embeddings.shape[-1]])

    def setup_from_data(self, data: Iterable[Tuple[QASetting, List[Answer]]]):
        vocab = self.shared_resources.vocab
        if not vocab.frozen:
            preprocessing.fill_vocab(
                (q for q, _ in data), vocab, lowercase=self.shared_resources.config.get('lowercase', True))
            vocab.freeze()
        if not hasattr(self.shared_resources, 'answer_vocab') or not self.shared_resources.answer_vocab.frozen:
            self.shared_resources.answer_vocab = util.create_answer_vocab(
                qa_settings=(q for q, _ in data), answers=(a for _, ass in data for a in ass))
            self.shared_resources.answer_vocab.freeze()
        self.shared_resources.char_vocab = preprocessing.char_vocab_from_vocab(self.shared_resources.vocab)

    @property
    def training_ports(self) -> List[TensorPort]:
        return [Ports.Target.target_index]

    @property
    def output_ports(self) -> List[TensorPort]:
        """Defines the outputs of the InputModule"""
        if self.shared_resources.embeddings is not None:
            return [Ports.Input.emb_support, Ports.Input.emb_question,
                    Ports.Input.support, Ports.Input.question,
                    Ports.Input.support_length, Ports.Input.question_length,
                    Ports.Input.sample_id,
                    # character information
                    Ports.Input.word_chars, Ports.Input.word_char_length,
                    Ports.Input.question_batch_words, Ports.Input.support_batch_words,
                    Ports.is_eval]
        else:
            return [Ports.Input.support, Ports.Input.question,
                    Ports.Input.support_length, Ports.Input.question_length,
                    Ports.Input.sample_id,
                    # character information
                    Ports.Input.word_chars, Ports.Input.word_char_length,
                    Ports.Input.question_batch_words, Ports.Input.support_batch_words,
                    Ports.is_eval]

    def preprocess(self, questions: List[QASetting],
                   answers: Optional[List[List[Answer]]] = None,
                   is_eval: bool = False) -> List[MCAnnotation]:
        if answers is None:
            answers = [None] * len(questions)
        preprocessed = []
        if len(questions) > 1000:
            bar = progressbar.ProgressBar(
                max_value=len(questions),
                widgets=[' [', progressbar.Timer(), '] ', progressbar.Bar(), ' (', progressbar.ETA(), ') '])
            for i, (q, a) in bar(enumerate(zip(questions, answers))):
                preprocessed.append(self.preprocess_instance(i, q, a))
        else:
            for i, (q, a) in enumerate(zip(questions, answers)):
                preprocessed.append(self.preprocess_instance(i, q, a))

        return preprocessed

    def preprocess_instance(self, idd: int, question: QASetting,
                            answers: Optional[List[Answer]] = None) -> MCAnnotation:
        has_answers = answers is not None

        q_tokenized, q_ids, q_length, _, _ = preprocessing.nlp_preprocess(
            question.question, self.shared_resources.vocab,
            lowercase=self.shared_resources.config.get('lowercase', True))
        s_tokenized, s_ids, s_length, _, _ = preprocessing.nlp_preprocess(
            question.support[0], self.shared_resources.vocab,
            lowercase=self.shared_resources.config.get('lowercase', True))

        return MCAnnotation(
            question_tokens=q_tokenized,
            question_ids=q_ids,
            question_length=q_length,
            support_tokens=s_tokenized,
            support_ids=s_ids,
            support_length=s_length,
            answer=self.shared_resources.answer_vocab(answers[0].text) if has_answers else 0,
            id=idd
        )

    def create_batch(self, annotations: List[MCAnnotation],
                     is_eval: bool, with_answers: bool) -> Mapping[TensorPort, np.ndarray]:
        # also add character information
        word_chars, word_lengths, tokens, vocab, rev_vocab = \
            preprocessing.unique_words_with_chars(
                [a.question_tokens for a in annotations] + [a.support_tokens for a in annotations],
                self.shared_resources.char_vocab)
        question_words, support_words = tokens[:len(annotations)], tokens[len(annotations):]

        q_lengths = [a.question_length for a in annotations]
        s_lengths = [a.support_length for a in annotations]
        xy_dict = {
            Ports.Input.question_length: q_lengths,
            Ports.Input.support_length: s_lengths,
            Ports.Input.sample_id: [a.id for a in annotations],
            Ports.Input.word_chars: word_chars,
            Ports.Input.word_char_length: word_lengths,
            Ports.Input.question_batch_words: question_words,
            Ports.Input.support_batch_words: support_words,
            Ports.is_eval: is_eval,
            Ports.Input.support: [a.support_ids for a in annotations],
            Ports.Input.question: [a.question_ids for a in annotations]
        }

        if self.embeddings is not None:
            emb_support = np.zeros([len(annotations), max(s_lengths), self.embeddings.shape[-1]])
            emb_question = np.zeros([len(annotations), max(q_lengths), self.embeddings.shape[-1]])
            for i, a in enumerate(annotations):
                for j, t in enumerate(a.support_tokens):
                    emb_support[i, j] = self.embeddings.get(t, self.__default_vec)
                for j, t in enumerate(a.question_tokens):
                    emb_question[i, j] = self.embeddings.get(t, self.__default_vec)

            xy_dict[Ports.Input.emb_support] = emb_support
            xy_dict[Ports.Input.emb_question] = emb_question

        if with_answers:
            xy_dict[Ports.Target.target_index] = [a.answer for a in annotations]
        return numpify(xy_dict)


def _np_softmax(x):
    """Compute softmax values for each sets of scores in x."""
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum(axis=0)


class SimpleClassificationOutputModule(OutputModule):
    def __init__(self, shared_resources=None):
        self._shared_resources = shared_resources

    def setup(self):
        pass

    @property
    def input_ports(self) -> List[TensorPort]:
        return [Ports.Prediction.logits]

    def __call__(self, questions: List[QASetting], tensors: Mapping[TensorPort, np.ndarray]) -> List[Answer]:
        # len(inputs) == batch size
        # logits: [batch_size, max_num_candidates]
        logits = tensors[Ports.Prediction.logits]
        winning_indices = np.argmax(logits, axis=1)
        result = []
        for index_in_batch, question in enumerate(questions):
            winning_index = winning_indices[index_in_batch]
            score = _np_softmax(logits[index_in_batch])[winning_index]
            if self._shared_resources is not None and hasattr(self._shared_resources, 'answer_vocab'):
                ans = Answer(self._shared_resources.answer_vocab.get_sym(winning_index), score=score)
            else:
                ans = Answer(question.candidates[winning_index], score=score)
            result.append([ans])
        return result



# -*- coding: utf-8 -*-

import logging

import numpy as np

logger = logging.getLogger(__name__)


def load_glove(stream, vocab=None):
    """Loads GloVe file and merges it if optional vocabulary
    Args:
        stream (iterable): An opened filestream to the GloVe file.
        vocab (dict=None): Word2idx dict of existing vocabulary.
    Returns:
        return_vocab (Vocabulary), lookup (matrix); Vocabulary contains the
                     word2idx and the matrix contains the embedded words.
    """
    logger.info('Loading GloVe vectors ..')

    word2idx = {}
    first_line = stream.readline()
    dim = len(first_line.split()) - 1
    lookup = np.empty([500000, dim], dtype=np.float)
    lookup[0] = np.fromstring(first_line.split(maxsplit=1)[1], sep=' ')
    word2idx[first_line.split(maxsplit=1)[0].decode('utf-8')] = 0
    n = 1
    for line in stream:
        word, vec = line.rstrip().split(maxsplit=1)
        if vocab is None or word in vocab and word not in word2idx:
            word = word.decode('utf-8')
            idx = len(word2idx)
            word2idx[word] = idx
            if idx > np.size(lookup, axis=0) - 1:
                lookup.resize([lookup.shape[0] + 500000, lookup.shape[1]])
            lookup[idx] = np.fromstring(vec, sep=' ')
        n += 1
    lookup.resize([len(word2idx), dim])
    logger.info('Loading GloVe vectors completed.')
    return word2idx, lookup



# -*- coding: utf-8 -*-

import tensorflow as tf


def parametric_relu(x, name=None):
    alphas = tf.get_variable('{}/alpha'.format(name) if name else 'alpha',
                             x.get_shape()[-1],
                             initializer=tf.constant_initializer(0.0),
                             dtype=tf.float32)
    return tf.nn.relu(x) + alphas * (x - abs(x)) * 0.5


def selu(x, name=None):
    with tf.name_scope('{}/elu'.format(name) if name else 'elu') as _:
        alpha = 1.6732632423543772848170429916717
        scale = 1.0507009873554804934193349852946
        return scale*tf.where(x >= 0.0, x, alpha*tf.nn.elu(x))


# Aliases
prelu = parametric_relu


def activation_from_string(activation_str):
    if activation_str is None:
        return tf.identity
    return getattr(tf.nn, activation_str)



"""

Richardson, Matthew, Christopher JC Burges, and Erin Renshaw.
"MCTest: A Challenge Dataset for the Open-Domain Machine Comprehension of Text."
EMNLP. Vol. 3. 2013.

Original paper: http://research.microsoft.com/en-us/um/redmond/projects/mctest/MCTest_EMNLP2013.pdf
Data:   research.microsoft.com/en-us/um/redmond/projects/mctest/
JTR download script: data/MCTest/download.sh

"""

import json
import argparse

labels = ["A", "B", "C", "D"]


def __clean_mctest_text(text):
    return text.replace('\\newline', '  ')


def create_jtr_snippet(tsv_file, ans_file, n_instances=5):
    return convert_mctest(tsv_file, ans_file, n_instances)


def convert_mctest(tsv_file, ans_file, first_n=None):
    with open(tsv_file) as tsv_data:
        tsv_lines = tsv_data.readlines()
    with open(ans_file) as ans_data:
        ans_lines = ans_data.readlines()
        instances = []
    for tsv, ans in zip(tsv_lines, ans_lines):
        instances.append(__parse_mctest_instance(tsv, ans))
    if first_n:
        instances = instances[:first_n]
    return {
        "meta": "MCTest",
        "instances": instances
    }



def __parse_mctest_instance(tsv_chunk, ans_chunk):
    tsv_tabs = tsv_chunk.strip().split('\t')
    ans_tabs = ans_chunk.strip().split('\t')
    # id = tsv_tabs[0]
    # ann = tsv_tabs[1]
    passage = tsv_tabs[2]
    # the dictionary for populating a set of passage/questions/answers
    qset_dict = {
        'support': [{'text': __clean_mctest_text(passage)}],
        'questions': __parse_mctest_questions(tsv_tabs[3:], ans_tabs)
    }
    return qset_dict


def __parse_mctest_questions(question_list, ans_tabs):
    # print(ans_tabs)
    questions = []
    for i in range(0, len(question_list), 5):
        # qdict = {}
        # parse answers
        candidates = []
        correct_answer = ans_tabs[int(i / 5)]
        for j in range(1, 5):
            label = labels[j-1]
            answer = {
                'label': label,
                'text': question_list[i + j]
            }
            candidates.append(answer)
        correct_index = labels.index(correct_answer)
        answer = {
            'index': correct_index,
            'text': question_list[i + correct_index + 1]
        }
        # parse question
        qcols = question_list[i].split(':')
        qdict = {
            'question': qcols[1],
            'candidates': candidates,
            'answers': [answer]
        }
        questions.append(qdict)
    return questions


def main():
    parser = argparse.ArgumentParser(description='Machine Comprehension Test (MCTest) dataset to jack format converter.')
    parser.add_argument('in_tsv',
                        help="path to the MCTest tsv file (e.g. data/MCTest/MCTest/mc160.train.tsv)")
    parser.add_argument('in_ans',
                        help="path to the MCTest ans file (e.g. data/MCTest/MCTest/mc160.train.ans)")
    parser.add_argument('outfile',
                        help="path to the jack format -generated output file (e.g. data/MCTest/train.160.jack.json)")
    parser.add_argument('-s', '--snippet', action="store_true",
                        help="Export a snippet (first 5 instances) instead of the full file")
    args = parser.parse_args()


    if args.snippet:
        corpus = convert_mctest(args.in_tsv, args.in_ans, 5)
    else:
        corpus = convert_mctest(args.in_tsv, args.in_ans)

    with open(args.outfile, 'w') as outfile:
        json.dump(corpus, outfile, indent=2)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""
Here we define light data structures to store the input to jack readers, and their output.
"""

from typing import Tuple, Sequence


class Answer:
    """
    Representation of an answer to a question.
    """

    def __init__(self, text: str, span: Tuple[int, int] = None, doc_idx: int = 0, score: float = 1.0):
        """
        Create a new answer.
        Args:
            text: The text string of the answer.
            span: For extractive QA, a span in the support documents. The triple `(start, end)`
                represents a span in support document with index `doc_index` in the ordered sequence of
            doc_idx: index of document where answer was found
            support documents. The span starts at `start` and ends at `end` (exclusive).
            score: the score a model associates with this answer.
        """
        assert span is None or len(span) == 2, "span should be (char_start, char_end) tuple"

        self.score = score
        self.span = span
        self.doc_idx = doc_idx
        self.text = text


class QASetting:
    """
    Representation of a single question answering problem. It primarily consists of a question,
    a list of support documents, and optionally, some set of candidate answers.
    """

    def __init__(self,
                 question: str,
                 support: Sequence[str] = (),
                 id: str = None,
                 candidates: Sequence[str] = None,
                 seq_candidates: Sequence[str] = None,
                 candidate_spans: Sequence[Tuple[int, int, int]] = None):
        """
        Create a new QASetting.
        Args:
            question: the question text.
            support: a sequence of support documents the answerer has access to when answering the question.
            id: an identifier for this question setting.
            candidates: a list of candidate answer strings.
            candidate_spans: for extractive QA, a sequence of candidate spans in the support documents.
            A span `(doc_index,start,end)` corresponds to a span in support document with index `doc_index`,
            with start position `start` and end position `end`.
        """
        self.id = id
        self.candidate_spans = candidate_spans
        self.candidates = candidates
        self.support = support
        self.question = question


def _jack_to_qasetting(instance, value, global_candidates):
    support = [value(s) for s in instance["support"]] if "support" in instance else None
    idd = value(instance, 'id')
    for question_instance in instance["questions"]:
        question = value(question_instance['question'])
        idd = value(question_instance, 'id') or idd
        idd = value(question_instance['question'], 'id') or idd
        if global_candidates is None:
            candidates = [value(c) for c in question_instance['candidates']] if "candidates" in question_instance else None
        else:
            candidates = global_candidates
        answers = [Answer(value(c), value(c, 'span'), value(c, 'doc_idx', 0)) for c in
                   question_instance['answers']] if "answers" in question_instance else None
        yield QASetting(question, support, candidates=candidates, id=idd), answers


def jack_to_qasetting(jtr_data, max_count=None):
    """
    Converts a python dictionary in Jack format to a QASetting.
    Args:
        jtr_data: dictionary extracted from jack json file.
        max_count: maximal number of instances to load.

    Returns:
        list of QASetting
    """

    def value(c, key="text", default=None):
        return c.get(key, default) if isinstance(c, dict) else c if key == 'text' else default

    global_candidates = [value(c) for c in jtr_data['globals']['candidates']] if 'globals' in jtr_data else None

    ans = [(inp, answer) for i in jtr_data["instances"]
           for inp, answer in _jack_to_qasetting(i, value, global_candidates)][:max_count]
    return ans



"""
This file contains FastQA specific modules and ports
"""

from jack.core import *
from jack.readers.extractive_qa.tensorflow.abstract_model import AbstractXQAModelModule
from jack.readers.extractive_qa.tensorflow.answer_layer import conditional_answer_layer, bilinear_answer_layer
from jack.util.tf import misc
from jack.util.tf.embedding import conv_char_embedding
from jack.util.tf.highway import highway_network
from jack.util.tf.sequence_encoder import encoder


class FastQAModule(AbstractXQAModelModule):
    def set_topk(self, k):
        self._topk_assign(k)

    def create_output(self, shared_resources, input_tensors):
        tensors = TensorPortTensors(input_tensors)
        with tf.variable_scope("fast_qa", initializer=tf.contrib.layers.xavier_initializer()):
            # Some helpers
            batch_size = tf.shape(tensors.question_length)[0]
            max_question_length = tf.reduce_max(tensors.question_length)
            support_mask = misc.mask_for_lengths(tensors.support_length)

            input_size = shared_resources.embeddings.shape[-1]
            size = shared_resources.config["repr_dim"]
            with_char_embeddings = shared_resources.config.get("with_char_embeddings", False)

            # set shapes for inputs
            tensors.emb_question.set_shape([None, None, input_size])
            tensors.emb_support.set_shape([None, None, input_size])

            emb_question = tensors.emb_question
            emb_support = tensors.emb_support
            if with_char_embeddings:
                # compute combined embeddings
                [char_emb_question, char_emb_support] = conv_char_embedding(
                    len(shared_resources.char_vocab), size, tensors.word_chars, tensors.word_char_length,
                    [tensors.question_batch_words, tensors.support_batch_words])

                emb_question = tf.concat([emb_question, char_emb_question], 2)
                emb_support = tf.concat([emb_support, char_emb_support], 2)
                input_size += size

                # set shapes for inputs
                emb_question.set_shape([None, None, input_size])
                emb_support.set_shape([None, None, input_size])

            # compute encoder features
            question_features = tf.ones(tf.stack([batch_size, max_question_length, 2]))

            v_wiqw = tf.get_variable("v_wiq_w", [1, 1, input_size],
                                     initializer=tf.constant_initializer(1.0))

            wiq_w = tf.matmul(tf.gather(emb_question * v_wiqw, tensors.support2question), emb_support, adjoint_b=True)
            wiq_w = wiq_w + tf.expand_dims(support_mask, 1)

            question_binary_mask = tf.gather(tf.sequence_mask(tensors.question_length, dtype=tf.float32),
                                             tensors.support2question)
            wiq_w = tf.reduce_sum(tf.nn.softmax(wiq_w) * tf.expand_dims(question_binary_mask, 2), [1])

            # [B, L , 2]
            support_features = tf.stack([tensors.word_in_question, wiq_w], 2)

            # highway layer to allow for interaction between concatenated embeddings
            if with_char_embeddings:
                with tf.variable_scope("char_embeddings") as vs:
                    emb_question = tf.layers.dense(emb_question, size, name="embeddings_projection")
                    emb_question = highway_network(emb_question, 1)
                    vs.reuse_variables()
                    emb_support = tf.layers.dense(emb_support, size, name="embeddings_projection")
                    emb_support = highway_network(emb_support, 1)

            keep_prob = 1.0 - shared_resources.config.get("dropout", 0.0)
            emb_question, emb_support = tf.cond(
                tensors.is_eval,
                lambda: (emb_question, emb_support),
                lambda: (tf.nn.dropout(emb_question, keep_prob, noise_shape=[1, 1, emb_question.get_shape()[-1].value]),
                         tf.nn.dropout(emb_support, keep_prob, noise_shape=[1, 1, emb_question.get_shape()[-1].value]))
            )

            # extend embeddings with features
            emb_question_ext = tf.concat([emb_question, question_features], 2)
            emb_support_ext = tf.concat([emb_support, support_features], 2)

            # encode question and support
            encoder_type = shared_resources.config.get('encoder', 'lstm').lower()
            if encoder_type in ['lstm', 'sru', 'gru']:
                size = size + 2 if encoder_type == 'sru' else size  # to allow for use of residual in SRU
                encoded_question = encoder(emb_question_ext, tensors.question_length, size, module=encoder_type)
                encoded_support = encoder(emb_support_ext, tensors.support_length, size, module=encoder_type,
                                          reuse=True)
                projection_initializer = tf.constant_initializer(np.concatenate([np.eye(size), np.eye(size)]))
                encoded_question = tf.layers.dense(encoded_question, size, tf.tanh, use_bias=False,
                                                   kernel_initializer=projection_initializer,
                                                   name='projection_q')
                encoded_support = tf.layers.dense(encoded_support, size, tf.tanh, use_bias=False,
                                                  kernel_initializer=projection_initializer, name='projection_s')
            else:
                raise ValueError("Only rnn ('lstm', 'sru', 'gru') encoder allowed for FastQA!")

            answer_layer = shared_resources.config.get('answer_layer', 'conditional').lower()

            topk = tf.get_variable(
                'topk', initializer=shared_resources.config.get('topk', 1), dtype=tf.int32, trainable=False)
            topk_p = tf.placeholder(tf.int32, [], 'beam_size_setter')
            topk_assign = topk.assign(topk_p)
            self._topk_assign = lambda k: self.tf_session.run(topk_assign, {topk_p: k})

            if answer_layer == 'conditional':
                start_scores, end_scores, doc_idx, predicted_start_pointer, predicted_end_pointer = \
                    conditional_answer_layer(size, encoded_question, tensors.question_length, encoded_support,
                                             tensors.support_length,
                                             tensors.correct_start, tensors.support2question, tensors.answer2support,
                                             tensors.is_eval,
                                             topk=topk,
                                             max_span_size=shared_resources.config.get("max_span_size", 10000))
            elif answer_layer == 'conditional_bilinear':
                start_scores, end_scores, doc_idx, predicted_start_pointer, predicted_end_pointer = \
                    conditional_answer_layer(size, encoded_question, tensors.question_length, encoded_support,
                                             tensors.support_length,
                                             tensors.correct_start, tensors.support2question, tensors.answer2support,
                                             tensors.is_eval,
                                             topk=topk,
                                             max_span_size=shared_resources.config.get("max_span_size", 10000),
                                             bilinear=True)
            elif answer_layer == 'bilinear':
                start_scores, end_scores, doc_idx, predicted_start_pointer, predicted_end_pointer = \
                    bilinear_answer_layer(size, encoded_question, tensors.question_length, encoded_support,
                                          tensors.support_length,
                                          tensors.support2question, tensors.answer2support, tensors.is_eval,
                                          topk=topk,
                                          max_span_size=shared_resources.config.get("max_span_size", 10000))
            else:
                raise ValueError

            span = tf.stack([doc_idx, predicted_start_pointer, predicted_end_pointer], 1)

            return TensorPort.to_mapping(self.output_ports, (start_scores, end_scores, span))



# -*- coding: utf-8 -*-

import tensorflow as tf


def mask_for_lengths(lengths, max_length=None, mask_right=True, value=-1000.0):
    """
    Creates a [batch_size x max_length] mask.

    Args:
        lengths: int32 1-dim tensor of batch_size lengths
        max_length: int32 0-dim tensor or python int
        mask_right: if True, everything before "lengths" becomes zero and the
            rest "value", else vice versa
        value: value for the mask

    Returns:
        [batch_size x max_length] mask of zeros and "value"s
    """
    mask = tf.sequence_mask(lengths, max_length, dtype=tf.float32)
    if mask_right:
        mask = 1.0 - mask
    mask *= value
    return mask



"""Shared utilities for multiple choice."""
from typing import Iterable

from jack.core.data_structures import QASetting, Answer
from jack.util.vocab import Vocab


def create_answer_vocab(qa_settings: Iterable[QASetting] = None, answers: Iterable[Answer] = None):
    vocab = Vocab(unk=None)
    if qa_settings is not None:
        for qa in qa_settings:
            if qa.candidates:
                for c in qa.candidates:
                    vocab(c)
    if answers is not None:
        for a in answers:
            vocab(a.text)
    return vocab


def candidate_one_hot(candidates, answer_str):
    return [1.0 if candidates[answer_str] == cand else 0.0 for cand in candidates]



# -*- coding: utf-8 -*-

import logging
from abc import abstractmethod

import numpy as np
import tensorflow as tf

from jack.readers.classification.shared import AbstractSingleSupportClassificationModel
from jack.util.tf.attention import attention_softmax3d
from jack.util.tf.masking import mask_3d

logger = logging.getLogger(__name__)


class DecomposableAttentionModel(AbstractSingleSupportClassificationModel):

    def forward_pass(self, shared_resources, embedded_question, embedded_support, num_classes, tensors,
                     has_bos_token=True):
        # final states_fw_bw dimensions:
        # [[[batch, output dim], [batch, output_dim]]

        if has_bos_token:
            # batch_size = embedded_question.get_shape()[0]
            embedding_size = embedded_question.get_shape().as_list()[2]

            bos_token_emb = tf.get_variable('bos_token_embedding',
                                            shape=(1, 1, embedding_size),
                                            initializer=tf.ones_initializer())

            batch_size = tf.shape(embedded_question)[0]

            t_bos_token_emb = tf.tile(
                input=bos_token_emb,
                multiples=[batch_size, 1, 1])

            embedded_question = tf.concat(values=[t_bos_token_emb, embedded_question], axis=1)
            embedded_support = tf.concat(values=[t_bos_token_emb, embedded_support], axis=1)

            tensors.question_length += 1
            tensors.support_length += 1

        if shared_resources.config.get('normalize_embeddings', False):
            embedded_question = tf.nn.l2_normalize(embedded_question, 2)
            embedded_support = tf.nn.l2_normalize(embedded_support, 2)

        dropout_rate = shared_resources.config.get('dropout', 0)
        dropout_keep_prob = tf.cond(tf.logical_not(tensors.is_eval), lambda: 1.0 - dropout_rate, lambda: 1.0)

        model_kwargs = {
            'sequence1': embedded_question,
            'sequence1_length': tensors.question_length,
            'sequence2': embedded_support,
            'sequence2_length': tensors.support_length,
            'representation_size': shared_resources.config['repr_dim'],
            'dropout_keep_prob': dropout_keep_prob,
            'use_masking': True,
        }

        model = FeedForwardDAM(**model_kwargs)
        logits = model()
        return logits


class BaseDecomposableAttentionModel:
    @abstractmethod
    def _transform_input(self, sequence, reuse=False):
        raise NotImplementedError

    @abstractmethod
    def _transform_attend(self, sequence, reuse=False):
        raise NotImplementedError

    @abstractmethod
    def _transform_compare(self, sequence, reuse=False):
        raise NotImplementedError

    @abstractmethod
    def _transform_aggregate(self, v1_v2, reuse=False):
        raise NotImplementedError

    def __init__(self, sequence1, sequence1_length, sequence2, sequence2_length,
                 nb_classes=3, reuse=False, use_masking=True, init_std_dev=0.01, *args, **kwargs):
        self.init_std_dev = init_std_dev
        self.nb_classes = nb_classes

        self.sequence1 = sequence1
        self.sequence1_length = sequence1_length

        self.sequence2 = sequence2
        self.sequence2_length = sequence2_length

        self.reuse = reuse

        embedding1_size = self.sequence1.get_shape()[-1].value
        embedding2_size = self.sequence2.get_shape()[-1].value

        assert embedding1_size == embedding2_size

        # [batch_size, time_steps, embedding_size] -> [batch_size, time_steps, representation_size]
        self.transformed_sequence1 = self._transform_input(self.sequence1, reuse=self.reuse)

        # [batch_size, time_steps, embedding_size] -> [batch_size, time_steps, representation_size]
        self.transformed_sequence2 = self._transform_input(self.sequence2, reuse=True)

        self.transformed_sequence1_length = self.sequence1_length
        self.transformed_sequence2_length = self.sequence2_length

        logger.info('Building the Attend graph ..')

        self.raw_attentions = None
        self.attention_sentence1 = self.attention_sentence2 = None

        # tensors with shape (batch_size, time_steps, num_units)
        self.alpha, self.beta = self.attend(self.transformed_sequence1, self.transformed_sequence2,
                                            sequence1_lengths=self.transformed_sequence1_length,
                                            sequence2_lengths=self.transformed_sequence2_length,
                                            use_masking=use_masking, reuse=self.reuse)

        logger.info('Building the Compare graph ..')

        # tensor with shape (batch_size, time_steps, num_units)
        self.v1 = self.compare(self.transformed_sequence1, self.beta, reuse=self.reuse)

        # tensor with shape (batch_size, time_steps, num_units)
        self.v2 = self.compare(self.transformed_sequence2, self.alpha, reuse=True)

        logger.info('Building the Aggregate graph ..')
        self.logits = self.aggregate(self.v1, self.v2, self.nb_classes,
                                     v1_lengths=self.transformed_sequence1_length,
                                     v2_lengths=self.transformed_sequence2_length,
                                     use_masking=use_masking, reuse=self.reuse)

    def __call__(self):
            return self.logits

    def attend(self, sequence1, sequence2,
               sequence1_lengths=None, sequence2_lengths=None, use_masking=True, reuse=False):
        """
        Attend phase.

        Args:
            sequence1: tensor with shape (batch_size, time_steps, num_units)
            sequence2: tensor with shape (batch_size, time_steps, num_units)
            sequence1_lengths: time_steps in sequence1
            sequence2_lengths: time_steps in sequence2
            use_masking: use masking
            reuse: reuse variables

        Returns:
            two tensors with shape (batch_size, time_steps, num_units)
        """
        with tf.variable_scope('attend') as _:
            # tensor with shape (batch_size, time_steps, num_units)
            transformed_sequence1 = self._transform_attend(sequence1, reuse)

            # tensor with shape (batch_size, time_steps, num_units)
            transformed_sequence2 = self._transform_attend(sequence2, True)

            # tensor with shape (batch_size, time_steps, time_steps)
            self.raw_attentions = tf.matmul(transformed_sequence1, tf.transpose(transformed_sequence2, [0, 2, 1]))

            masked_raw_attentions = self.raw_attentions
            if use_masking:
                masked_raw_attentions = mask_3d(sequences=masked_raw_attentions,
                                                sequence_lengths=sequence2_lengths,
                                                mask_value=- np.inf, dimension=2)
            self.attention_sentence1 = attention_softmax3d(masked_raw_attentions)

            # tensor with shape (batch_size, time_steps, time_steps)
            attention_transposed = tf.transpose(self.raw_attentions, [0, 2, 1])
            masked_attention_transposed = attention_transposed
            if use_masking:
                masked_attention_transposed = mask_3d(sequences=masked_attention_transposed,
                                                      sequence_lengths=sequence1_lengths,
                                                      mask_value=- np.inf, dimension=2)
            self.attention_sentence2 = attention_softmax3d(masked_attention_transposed)

            # tensors with shape (batch_size, time_steps, num_units)
            alpha = tf.matmul(self.attention_sentence2, sequence1, name='alpha')
            beta = tf.matmul(self.attention_sentence1, sequence2, name='beta')
            return alpha, beta

    def compare(self, sentence, soft_alignment, reuse=False):
        """
        Compare phase.

        Args:
            sentence: tensor with shape (batch_size, time_steps, num_units)
            soft_alignment: tensor with shape (batch_size, time_steps, num_units)
            reuse: reuse variables

        Returns:
            tensor with shape (batch_size, time_steps, num_units)
        """
        # tensor with shape (batch, time_steps, num_units)
        sentence_and_alignment = tf.concat(axis=2, values=[sentence, soft_alignment])
        transformed_sentence_and_alignment = self._transform_compare(sentence_and_alignment, reuse=reuse)
        return transformed_sentence_and_alignment

    def aggregate(self, v1, v2, num_classes,
                  v1_lengths=None, v2_lengths=None, use_masking=True, reuse=False):
        """
        Aggregate phase.

        Args:
            v1: tensor with shape (batch_size, time_steps, num_units)
            v2: tensor with shape (batch_size, time_steps, num_units)
            num_classes: number of output units
            v1_lengths: time_steps in v1
            v2_lengths: time_steps in v2
            use_masking: use masking
            reuse: reuse variables
        """
        with tf.variable_scope('aggregate', reuse=reuse) as _:
            if use_masking:
                v1 = mask_3d(sequences=v1, sequence_lengths=v1_lengths, mask_value=0, dimension=1)
                v2 = mask_3d(sequences=v2, sequence_lengths=v2_lengths, mask_value=0, dimension=1)

            v1_sum, v2_sum = tf.reduce_sum(v1, [1]), tf.reduce_sum(v2, [1])

            v1_v2 = tf.concat(axis=1, values=[v1_sum, v2_sum])
            transformed_v1_v2 = self._transform_aggregate(v1_v2, reuse=reuse)

            logits = tf.contrib.layers.fully_connected(inputs=transformed_v1_v2,
                                                       num_outputs=num_classes,
                                                       weights_initializer=tf.random_normal_initializer(0.0, 0.01),
                                                       biases_initializer=tf.zeros_initializer(),
                                                       activation_fn=None)
        return logits


class FeedForwardDAM(BaseDecomposableAttentionModel):
    def __init__(self, representation_size=200, dropout_keep_prob=1.0, *args, **kwargs):
        self.representation_size = representation_size
        self.dropout_keep_prob = dropout_keep_prob
        super().__init__(*args, **kwargs)

    def _transform_input(self, sequence, reuse=False):
        with tf.variable_scope('transform_embeddings', reuse=reuse) as _:
            projection = tf.contrib.layers.fully_connected(inputs=sequence, num_outputs=self.representation_size,
                                                           weights_initializer=tf.random_normal_initializer(0.0,
                                                                                                            self.init_std_dev),
                                                           biases_initializer=None, activation_fn=None)
        return projection

    def _transform_attend(self, sequence, reuse=False):
        with tf.variable_scope('transform_attend', reuse=reuse) as _:
            projection = tf.nn.dropout(sequence, keep_prob=self.dropout_keep_prob)
            projection = tf.contrib.layers.fully_connected(inputs=projection, num_outputs=self.representation_size,
                                                           weights_initializer=tf.random_normal_initializer(0.0,
                                                                                                            self.init_std_dev),
                                                           biases_initializer=tf.zeros_initializer(),
                                                           activation_fn=tf.nn.relu)
            projection = tf.nn.dropout(projection, keep_prob=self.dropout_keep_prob)
            projection = tf.contrib.layers.fully_connected(inputs=projection, num_outputs=self.representation_size,
                                                           weights_initializer=tf.random_normal_initializer(0.0,
                                                                                                            self.init_std_dev),
                                                           biases_initializer=tf.zeros_initializer(),
                                                           activation_fn=tf.nn.relu)
        return projection

    def _transform_compare(self, sequence, reuse=False):
        with tf.variable_scope('transform_compare', reuse=reuse) as _:
            projection = tf.nn.dropout(sequence, keep_prob=self.dropout_keep_prob)
            projection = tf.contrib.layers.fully_connected(inputs=projection, num_outputs=self.representation_size,
                                                           weights_initializer=tf.random_normal_initializer(0.0,
                                                                                                            self.init_std_dev),
                                                           biases_initializer=tf.zeros_initializer(),
                                                           activation_fn=tf.nn.relu)
            projection = tf.nn.dropout(projection, keep_prob=self.dropout_keep_prob)
            projection = tf.contrib.layers.fully_connected(inputs=projection, num_outputs=self.representation_size,
                                                           weights_initializer=tf.random_normal_initializer(0.0,
                                                                                                            self.init_std_dev),
                                                           biases_initializer=tf.zeros_initializer(),
                                                           activation_fn=tf.nn.relu)
        return projection

    def _transform_aggregate(self, v1_v2, reuse=False):
        with tf.variable_scope('transform_aggregate', reuse=reuse) as _:
            projection = tf.nn.dropout(v1_v2, keep_prob=self.dropout_keep_prob)
            projection = tf.contrib.layers.fully_connected(inputs=projection, num_outputs=self.representation_size,
                                                           weights_initializer=tf.random_normal_initializer(0.0,
                                                                                                            self.init_std_dev),
                                                           biases_initializer=tf.zeros_initializer(),
                                                           activation_fn=tf.nn.relu)
            projection = tf.nn.dropout(projection, keep_prob=self.dropout_keep_prob)
            projection = tf.contrib.layers.fully_connected(inputs=projection, num_outputs=self.representation_size,
                                                           weights_initializer=tf.random_normal_initializer(0.0,
                                                                                                            self.init_std_dev),
                                                           biases_initializer=tf.zeros_initializer(),
                                                           activation_fn=tf.nn.relu)
        return projection
